@@ -13,7 +13,7 @@
 **	(c) Copyright CERN 1991 - See Copyright.html
 */
 
-#define HASH_SIZE 101		/* Arbitrary prime. Memory/speed tradeoff */
+#define HASH_SIZE 101		/* Arbitrary prime.  Memory/speed tradeoff */
 
 #include <HTUtils.h>
 #include <HTAnchor.h>
@@ -21,6 +21,7 @@
 #include <UCAux.h>
 #include <UCMap.h>
 
+#include <LYUtils.h>
 #include <LYCharSets.h>
 #include <LYLeaks.h>
 
@@ -60,7 +61,7 @@ PRIVATE HTList **adult_table = 0;  /* Point to table of lists of all parents */
 /*				Creation Methods
 **				================
 **
-**	Do not use "new" by itself outside this module. In order to enforce
+**	Do not use "new" by itself outside this module.  In order to enforce
 **	consistency, we insist that you furnish more information about the
 **	anchor you are creating : use newWithParent or newWithAddress.
 */
@@ -68,11 +69,17 @@ PRIVATE HTParentAnchor * HTParentAnchor_new NOARGS
 {
     HTParentAnchor *newAnchor =
        (HTParentAnchor *)calloc(1, sizeof(HTParentAnchor));  /* zero-filled */
+    if (newAnchor == NULL)
+	outofmem(__FILE__, "HTParentAnchor_new");
     newAnchor->parent = newAnchor;
     newAnchor->bookmark = NULL; 	/* Bookmark filename. - FM */
     newAnchor->isISMAPScript = FALSE;	/* Lynx appends ?0,0 if TRUE. - FM */
     newAnchor->isHEAD = FALSE;		/* HEAD request if TRUE. - FM */
     newAnchor->safe = FALSE;		/* Safe. - FM */
+#ifdef SOURCE_CACHE
+    newAnchor->source_cache_file = NULL;
+    newAnchor->source_cache_chunk = NULL;
+#endif
     newAnchor->FileCache = NULL;	/* Path to a disk-cached copy. - FM */
     newAnchor->SugFname = NULL; 	/* Suggested filename. - FM */
     newAnchor->RevTitle = NULL; 	/* TITLE for a LINK with REV. - FM */
@@ -89,13 +96,19 @@ PRIVATE HTParentAnchor * HTParentAnchor_new NOARGS
     newAnchor->date = NULL;		/* Date. - FM */
     newAnchor->expires = NULL;		/* Expires. - FM */
     newAnchor->last_modified = NULL;	/* Last-Modified. - FM */
+    newAnchor->ETag = NULL;		/* ETag (HTTP/1.1 cache validator) */
     newAnchor->server = NULL;		/* Server. - FM */
     return(newAnchor);
 }
 
 PRIVATE HTChildAnchor * HTChildAnchor_new NOARGS
 {
-    return (HTChildAnchor *)calloc(1, sizeof(HTChildAnchor)); /* zero-filled */
+    HTChildAnchor *p;
+
+    p = (HTChildAnchor *)calloc(1, sizeof(HTChildAnchor)); /* zero-filled */
+    if (p == NULL)
+	outofmem(__FILE__, "HTChildAnchor_new");
+    return p;
 }
 
 
@@ -141,14 +154,20 @@ PRIVATE BOOL HTIdentical ARGS2(
 	CONST char *,	t)
 {
     if (s && t) {  /* Make sure they point to something */
+#ifdef SH_EX	/* 1998/04/28 (Tue) 22:02:58 */
+	if (*s == 'P' || *t == 'P') {
+	    if (strcmp(s + 1, "Name") == 0 || strcmp(t + 1, "Name") == 0)
+		return NO;
+	}
+#endif
 	for (; *s && *t; s++, t++) {
 	    if (*s != *t) {
 		return(NO);
 	    }
 	}
-	return(*s == *t);
+	return (BOOL) (*s == *t);
     } else {
-	return(s == t); 	/* Two NULLs are identical, aren't they ? */
+	return (BOOL) (s == t);  /* Two NULLs are identical, aren't they ? */
     }
 }
 #endif /* CASE_INSENSITIVE_ANCHORS */
@@ -158,7 +177,7 @@ PRIVATE BOOL HTIdentical ARGS2(
 **	---------------------------------
 **
 **	Me one is for a new anchor being edited into an existing
-**	document. The parent anchor must already exist.
+**	document.  The parent anchor must already exist.
 */
 PUBLIC HTChildAnchor * HTAnchor_findChild ARGS2(
 	HTParentAnchor *,	parent,
@@ -168,7 +187,7 @@ PUBLIC HTChildAnchor * HTAnchor_findChild ARGS2(
     HTList *kids;
 
     if (!parent) {
-	CTRACE(tfp, "HTAnchor_findChild called with NULL parent.\n");
+	CTRACE((tfp, "HTAnchor_findChild called with NULL parent.\n"));
 	return(NULL);
     }
     if ((kids = parent->children) != 0) {
@@ -183,8 +202,8 @@ PUBLIC HTChildAnchor * HTAnchor_findChild ARGS2(
 		if (HTIdentical(child->tag, tag)) /* Case sensitive - FM */
 #endif /* CASE_INSENSITIVE_ANCHORS */
 		{
-		    CTRACE(tfp, "Child anchor %p of parent %p with name `%s' already exists.\n",
-				(void *)child, (void *)parent, tag);
+		    CTRACE((tfp, "Child anchor %p of parent %p with name `%s' already exists.\n",
+				(void *)child, (void *)parent, tag));
 		    return(child);
 		}
 	    }
@@ -194,10 +213,10 @@ PUBLIC HTChildAnchor * HTAnchor_findChild ARGS2(
     }
 
     child = HTChildAnchor_new();
-    CTRACE(tfp, "new Anchor %p named `%s' is child of %p\n",
+    CTRACE((tfp, "HTAnchor: New Anchor %p named `%s' is child of %p\n",
 		(void *)child,
 		tag ? tag : (CONST char *)"",
-		(void *)parent); /* int for apollo */
+		(void *)parent)); /* int for apollo */
     HTList_addObject (parent->children, child);
     child->parent = parent;
     StrAllocCopy(child->tag, tag);
@@ -220,7 +239,7 @@ PUBLIC HTChildAnchor * HTAnchor_findChildAndLink ARGS4(
 {
     HTChildAnchor * child = HTAnchor_findChild(parent, tag);
 
-    CTRACE(tfp,"Entered HTAnchor_findChildAndLink\n");
+    CTRACE((tfp,"Entered HTAnchor_findChildAndLink\n"));
 
     if (href && *href) {
 	char *relative_to = HTAnchor_address((HTAnchor *)parent);
@@ -245,6 +264,29 @@ PUBLIC HTChildAnchor * HTAnchor_findChildAndLink ARGS4(
 	parsed_doc.safe = FALSE;
 	dest = HTAnchor_findAddress(&parsed_doc);
 
+#define DUPLICATE_ANCHOR_NAME_WORKAROUND
+
+#ifdef DUPLICATE_ANCHOR_NAME_WORKAROUND
+	if (tag && *tag) {
+	    HTAnchor *testdest1;
+	    int nlinks;
+	    testdest1 = child->mainLink.dest;
+	    if (testdest1) {
+		nlinks = 1 + HTList_count(child->links);
+		CTRACE((tfp,
+		       "*** Duplicate ChildAnchor %p named `%s' with %d links",
+		       child, tag, nlinks));
+		if (dest == testdest1 && ltype == child->mainLink.type) {
+		    CTRACE((tfp,", same dest %p and type, keeping it\n",
+			   testdest1));
+		} else {
+		    CTRACE((tfp,", different dest %p, creating unnamed child\n",
+			   testdest1));
+		    child = HTAnchor_findChild(parent, 0);
+		}
+	    }
+	}
+#endif
 	HTAnchor_link((HTAnchor *)child, dest, ltype);
 	FREE(parsed_doc.address);
 	FREE(relative_to);
@@ -252,6 +294,7 @@ PUBLIC HTChildAnchor * HTAnchor_findChildAndLink ARGS4(
     return(child);
 }
 
+#ifdef LY_FIND_LEAKS
 /*
 **  Function for freeing the adult hash table. - FM
 */
@@ -275,6 +318,12 @@ PRIVATE void free_adult_table NOARGS
 	    adult_table[i_counter] = HTAp_freeme->next;
 	    if (HTAp_freeme->object) {
 		parent = (HTParentAnchor *)HTAp_freeme->object;
+		CTRACE((tfp, "delete anchor:%d/%d,%d,%d %s\n",
+		       i_counter, HTList_count(HTAp_freeme) + 1,
+		       (parent->physical ? 1 : 0),
+		       (int)parent->underway,
+		       (parent->address ? parent->address : "(no address)")));
+		parent->underway = FALSE;
 		HTAnchor_delete(parent);
 	    }
 	    FREE(HTAp_freeme);
@@ -282,6 +331,7 @@ PRIVATE void free_adult_table NOARGS
     }
     FREE(adult_table);
 }
+#endif /* LY_FIND_LEAKS */
 
 /*	Create new or find old named anchor
 **	-----------------------------------
@@ -297,7 +347,7 @@ PUBLIC HTAnchor * HTAnchor_findAddress ARGS1(
     /* Anchor tag specified ? */
     char *tag = HTParse(newdoc->address, "", PARSE_ANCHOR);
 
-    CTRACE(tfp,"Entered HTAnchor_findAddress\n");
+    CTRACE((tfp,"Entered HTAnchor_findAddress\n"));
 
     /*
     **	If the address represents a sub-anchor, we recursively load its
@@ -339,7 +389,11 @@ PUBLIC HTAnchor * HTAnchor_findAddress ARGS1(
 	hash = HASH_FUNCTION(newdoc->address);
 	if (!adult_table) {
 	    adult_table = (HTList **)calloc(HASH_SIZE, sizeof(HTList *));
+	    if (!adult_table)
+		outofmem(__FILE__, "HTAnchor_findAddress");
+#ifdef LY_FIND_LEAKS
 	    atexit(free_adult_table);
+#endif
 	}
 	if (!adult_table[hash])
 	    adult_table[hash] = HTList_new();
@@ -361,8 +415,8 @@ PUBLIC HTAnchor * HTAnchor_findAddress ARGS1(
 		foundAnchor->isHEAD == newdoc->isHEAD)
 #endif /* CASE_INSENSITIVE_ANCHORS */
 	    {
-		CTRACE(tfp, "Anchor %p with address `%s' already exists.\n",
-			    (void *)foundAnchor, newdoc->address);
+		CTRACE((tfp, "Anchor %p with address `%s' already exists.\n",
+			    (void *)foundAnchor, newdoc->address));
 		 return (HTAnchor *)foundAnchor;
 	     }
 	}
@@ -371,8 +425,8 @@ PUBLIC HTAnchor * HTAnchor_findAddress ARGS1(
 	**  Node not found: create new anchor.
 	*/
 	foundAnchor = HTParentAnchor_new();
-	CTRACE(tfp, "New anchor %p has hash %d and address `%s'\n",
-		    (void *)foundAnchor, hash, newdoc->address);
+	CTRACE((tfp, "New anchor %p has hash %d and address `%s'\n",
+		    (void *)foundAnchor, hash, newdoc->address));
 	StrAllocCopy(foundAnchor->address, newdoc->address);
 	if (newdoc->post_data)
 	    StrAllocCopy(foundAnchor->post_data, newdoc->post_data);
@@ -387,7 +441,25 @@ PUBLIC HTAnchor * HTAnchor_findAddress ARGS1(
 	return (HTAnchor *)foundAnchor;
     }
 }
+/*	Create new or find old named anchor - simple form
+**	-------------------------------------------------
+**
+**	Like the previous one, but simpler to use for simple cases.
+**	No post data etc. can be supplied. - kw
+*/
+PUBLIC HTAnchor * HTAnchor_findSimpleAddress ARGS1(
+	CONST char *,	url)
+{
+    DocAddress urldoc;
 
+    urldoc.address = (char *)url; /* ignore warning, it IS treated like const - kw */
+    urldoc.post_data = NULL;
+    urldoc.post_content_type = NULL;
+    urldoc.bookmark = NULL;
+    urldoc.isHEAD = FALSE;
+    urldoc.safe = FALSE;
+    return HTAnchor_findAddress(&urldoc);
+}
 
 /*	Delete an anchor and possibly related things (auto garbage collection)
 **	--------------------------------------------
@@ -419,6 +491,14 @@ PRIVATE void deleteLinks ARGS1(
     if (me->mainLink.dest) {
 	HTParentAnchor *parent = me->mainLink.dest->parent;
 
+	/*
+	 *  Set the mainLink pointer to zero NOW.  If we don't,
+	 *  and we get somehow called recursively again for this
+	 *  same old me during the HTAnchor_delete below, weird
+	 *  things can occasionally happen. - kw
+	 */
+	 me->mainLink.dest = NULL;
+
 	/*
 	 *  Remove me from the parent's sources so that the
 	 *  parent knows one less anchor is it's dest.
@@ -434,7 +514,7 @@ PRIVATE void deleteLinks ARGS1(
 	 *  Test here to avoid calling overhead.
 	 *  If the parent has no loaded document, then we should
 	 *  tell it to attempt to delete itself.
-	 *  Don't do this jass if the anchor passed in is the same
+	 *  Don't do this jazz if the anchor passed in is the same
 	 *  as the anchor to delete.
 	 *  Also, don't do this if the destination parent is our
 	 *  parent.
@@ -451,7 +531,6 @@ PRIVATE void deleteLinks ARGS1(
 	 *  Leave the HTAtom pointed to by type up to other code to
 	 *  handle (reusable, near static).
 	 */
-	me->mainLink.dest = NULL;
 	me->mainLink.type = NULL;
     }
 
@@ -519,6 +598,28 @@ PRIVATE void deleteLinks ARGS1(
     }
 }
 
+#ifdef SOURCE_CACHE
+PUBLIC void HTAnchor_clearSourceCache ARGS1(
+	HTParentAnchor *,	me)
+{
+    /*
+     * Clean up the source cache, if any.
+     */
+    if (me->source_cache_file) {
+	CTRACE((tfp, "SourceCache: Removing file %s\n",
+	       me->source_cache_file));
+	LYRemoveTemp(me->source_cache_file);
+	FREE(me->source_cache_file);
+    }
+    if (me->source_cache_chunk) {
+	CTRACE((tfp, "SourceCache: Removing memory chunk %p\n",
+	       (void *)me->source_cache_chunk));
+	HTChunkFree(me->source_cache_chunk);
+	me->source_cache_chunk = NULL;
+    }
+}
+#endif /* SOURCE_CACHE */
+
 PUBLIC BOOL HTAnchor_delete ARGS1(
 	HTParentAnchor *,	me)
 {
@@ -543,6 +644,12 @@ PUBLIC BOOL HTAnchor_delete ARGS1(
 	return(NO);
     }
 
+    /*
+     *  Mark ourselves busy, so that recursive calls of this function
+     *  on this HTParentAnchor will not free it from under our feet. - kw
+     */
+    me->underway = TRUE;
+
     /*
      *	Recursively try to delete destination anchors of this parent.
      *	In any event, this will tell all destination anchors that we
@@ -569,6 +676,7 @@ PUBLIC BOOL HTAnchor_delete ARGS1(
 		}
 	    }
 	}
+	me->underway = FALSE;
 
 	/*
 	 *  Can't delete parent, still have sources.
@@ -592,6 +700,7 @@ PUBLIC BOOL HTAnchor_delete ARGS1(
 	    }
 	}
     }
+    me->underway = FALSE;
 
     /*
      *	Delete our empty list of children.
@@ -634,6 +743,9 @@ PUBLIC BOOL HTAnchor_delete ARGS1(
     FREE(me->bookmark);
     FREE(me->owner);
     FREE(me->RevTitle);
+#ifdef SOURCE_CACHE
+    HTAnchor_clearSourceCache(me);
+#endif
     if (me->FileCache) {
 	FILE *fd;
 	if ((fd = fopen(me->FileCache, "r")) != NULL) {
@@ -651,9 +763,12 @@ PUBLIC BOOL HTAnchor_delete ARGS1(
     FREE(me->content_disposition);
     FREE(me->content_location);
     FREE(me->content_md5);
+    FREE(me->message_id);
+    FREE(me->subject);
     FREE(me->date);
     FREE(me->expires);
     FREE(me->last_modified);
+    FREE(me->ETag);
     FREE(me->server);
 #ifdef USE_HASH
     FREE(me->style);
@@ -735,16 +850,6 @@ PUBLIC HyperDoc * HTAnchor_document ARGS1(
 }
 
 
-/*  We don't want code to change an address after anchor creation... yet ?
-PUBLIC void HTAnchor_setAddress ARGS2(
-	HTAnchor *,	me,
-	char *, 	addr)
-{
-    if (me)
-	StrAllocCopy (me->parent->address, addr);
-}
-*/
-
 PUBLIC char * HTAnchor_address ARGS1(
 	HTAnchor *,	me)
 {
@@ -755,13 +860,8 @@ PUBLIC char * HTAnchor_address ARGS1(
 	    !((HTChildAnchor *)me)->tag) {  /* it's an adult or no tag */
 	    StrAllocCopy(addr, me->parent->address);
 	} else {  /* it's a named child */
-	    addr = malloc(2 +
-			  strlen(me->parent->address) +
-			  strlen(((HTChildAnchor *)me)->tag));
-	    if (addr == NULL)
-		outofmem(__FILE__, "HTAnchor_address");
-	    sprintf(addr, "%s#%s",
-			   me->parent->address, ((HTChildAnchor *)me)->tag);
+	    HTSprintf0(&addr, "%s#%s",
+		       me->parent->address, ((HTChildAnchor *)me)->tag);
 	}
     }
     return(addr);
@@ -818,7 +918,7 @@ PUBLIC BOOL HTAnchor_isISMAPScript ARGS1(
 PUBLIC BOOL HTAnchor_hasChildren ARGS1(
 	HTParentAnchor *,	me)
 {
-    return( me ? ! HTList_isEmpty(me->children) : NO);
+    return (BOOL) ( me ? ! HTList_isEmpty(me->children) : NO);
 }
 
 #if defined(USE_HASH)
@@ -856,11 +956,21 @@ PUBLIC void HTAnchor_setTitle ARGS2(
     int i;
 
     if (me) {
-	StrAllocCopy(me->title, title);
-	for (i = 0; me->title[i]; i++) {
-	    if ((unsigned char)me->title[i] == 1 ||
-		(unsigned char)me->title[i] == 2) {
-		me->title[i] = ' ';
+	if (title) {
+	    StrAllocCopy(me->title, title);
+	    for (i = 0; me->title[i]; i++) {
+		if ((unsigned char)me->title[i] == 1 ||
+		    (unsigned char)me->title[i] == 2) {
+		    me->title[i] = ' ';
+		}
+	    }
+	} else {
+	    CTRACE((tfp,"HTAnchor_setTitle: New title is NULL! "));
+	    if (me->title) {
+		CTRACE((tfp,"Old title was \"%s\".\n", me->title));
+		FREE(me->title);
+	    } else {
+		CTRACE((tfp,"Old title was NULL.\n"));
 	    }
 	}
     }
@@ -998,7 +1108,7 @@ PUBLIC CONST char * HTAnchor_server ARGS1(
 PUBLIC BOOL HTAnchor_safe ARGS1(
 	HTParentAnchor *,	me)
 {
-    return( me ? me->safe : FALSE);
+    return (BOOL) ( me ? me->safe : FALSE);
 }
 
 /*	Content-Base header handling. - FM
@@ -1017,6 +1127,44 @@ PUBLIC CONST char * HTAnchor_content_location ARGS1(
     return( me ? me->content_location : NULL);
 }
 
+/*	Message-ID, used for mail replies - kw
+*/
+PUBLIC CONST char * HTAnchor_messageID ARGS1(
+	HTParentAnchor *,	me)
+{
+    return( me ? me->message_id : NULL);
+}
+
+PUBLIC BOOL HTAnchor_setMessageID ARGS2(
+	HTParentAnchor *,	me,
+	CONST char *,		messageid)
+{
+    if (!(me && messageid && *messageid)) {
+	return FALSE;
+    }
+    StrAllocCopy(me->message_id, messageid);
+    return TRUE;
+}
+
+/*	Subject, used for mail replies - kw
+*/
+PUBLIC CONST char * HTAnchor_subject ARGS1(
+	HTParentAnchor *,	me)
+{
+    return( me ? me->subject : NULL);
+}
+
+PUBLIC BOOL HTAnchor_setSubject ARGS2(
+	HTParentAnchor *,	me,
+	CONST char *,		subject)
+{
+    if (!(me && subject && *subject)) {
+	return FALSE;
+    }
+    StrAllocCopy(me->subject, subject);
+    return TRUE;
+}
+
 /*	Link me Anchor to another given one
 **	-------------------------------------
 */
@@ -1027,7 +1175,7 @@ PUBLIC BOOL HTAnchor_link ARGS3(
 {
     if (!(source && destination))
 	return(NO);  /* Can't link to/from non-existing anchor */
-    CTRACE(tfp, "Linking anchor %p to anchor %p\n", source, destination);
+    CTRACE((tfp, "Linking anchor %p to anchor %p\n", source, destination));
     if (!source->mainLink.dest) {
 	source->mainLink.dest = destination;
 	source->mainLink.type = type;
