@@ -10,50 +10,49 @@
 **
 */
 
-#include "HTUtils.h"
-#include "tcp.h"
-#include "HTAccess.h"
+#include <HTUtils.h>
 
 /* Implements:
 */
-#include "HTFormat.h"
+#include <HTFormat.h>
 
 PUBLIC float HTMaxSecs = 1e10;		/* No effective limit */
 PUBLIC float HTMaxLength = 1e10;	/* No effective limit */
 PUBLIC long int HTMaxBytes  = 0;	/* No effective limit */
 
-#ifdef unix
+#ifdef UNIX
 #ifdef NeXT
 #define PRESENT_POSTSCRIPT "open %s; /bin/rm -f %s\n"
 #else
 #define PRESENT_POSTSCRIPT "(ghostview %s ; /bin/rm -f %s)&\n"
 			   /* Full pathname would be better! */
 #endif /* NeXT */
-#endif /* unix */
+#endif /* UNIX */
 
-#include "HTML.h"
-#include "HTMLDTD.h"
-#include "HText.h"
-#include "HTAlert.h"
-#include "HTList.h"
-#include "HTInit.h"
-#include "HTTCP.h"
+#include <HTML.h>
+#include <HTMLDTD.h>
+#include <HText.h>
+#include <HTAlert.h>
+#include <HTList.h>
+#include <HTInit.h>
+#include <HTTCP.h>
+#include <HTTP.h>
 /*	Streams and structured streams which we use:
 */
-#include "HTFWriter.h"
-#include "HTPlain.h"
-#include "SGML.h"
-#include "HTML.h"
-#include "HTMLGen.h"
+#include <HTFWriter.h>
+#include <HTPlain.h>
+#include <SGML.h>
+#include <HTMLGen.h>
 
-#include "LYexit.h"
-#include "LYUtils.h"
-#include "LYGlobalDefs.h"
-#include "LYLeaks.h"
+#include <LYexit.h>
+#include <LYUtils.h>
+#include <GridText.h>
+#include <LYGlobalDefs.h>
+#include <LYLeaks.h>
 
-#define FREE(x) if (x) {free(x); x = NULL;}
-
-extern int HTCheckForInterrupt NOPARAMS;
+#ifdef DISP_PARTIAL
+#include <LYMainLoop.h>
+#endif
 
 PUBLIC	BOOL HTOutputSource = NO;	/* Flag: shortcut parser to stdout */
 /* extern  BOOL interactive; LJM */
@@ -81,7 +80,9 @@ PUBLIC	HTPresentation * default_presentation = NULL;
 /*
  *	To free off the presentation list.
  */
+#ifdef LY_FIND_LEAKS
 PRIVATE void HTFreePresentations NOPARAMS;
+#endif
 
 /*	Define a presentation system command for a content-type
 **	-------------------------------------------------------
@@ -89,9 +90,9 @@ PRIVATE void HTFreePresentations NOPARAMS;
 PUBLIC void HTSetPresentation ARGS6(
 	CONST char *,	representation,
 	CONST char *,	command,
-	float,		quality,
-	float,		secs,
-	float,		secs_per_byte,
+	double,		quality,
+	double,		secs,
+	double,		secs_per_byte,
 	long int,	maxbytes)
 {
     HTPresentation * pres = (HTPresentation *)malloc(sizeof(HTPresentation));
@@ -100,10 +101,10 @@ PUBLIC void HTSetPresentation ARGS6(
 
     pres->rep = HTAtom_for(representation);
     pres->rep_out = WWW_PRESENT;		/* Fixed for now ... :-) */
-    pres->converter = HTSaveAndExecute; 	/* Fixed for now ...	 */
-    pres->quality = quality;
-    pres->secs = secs;
-    pres->secs_per_byte = secs_per_byte;
+    pres->converter = HTSaveAndExecute;		/* Fixed for now ...	 */
+    pres->quality = (float) quality;
+    pres->secs = (float) secs;
+    pres->secs_per_byte = (float) secs_per_byte;
     pres->maxbytes = maxbytes;
     pres->command = NULL;
     StrAllocCopy(pres->command, command);
@@ -114,7 +115,9 @@ PUBLIC void HTSetPresentation ARGS6(
      */
     if (!HTPresentations)	{
 	HTPresentations = HTList_new();
+#ifdef LY_FIND_LEAKS
 	atexit(HTFreePresentations);
+#endif
     }
 
     if (strcmp(representation, "*")==0) {
@@ -157,12 +160,15 @@ PUBLIC void HTSetConversion ARGS7(
      */
     if (!HTPresentations)	{
 	HTPresentations = HTList_new();
+#ifdef LY_FIND_LEAKS
 	atexit(HTFreePresentations);
+#endif
     }
 
     HTList_addObject(HTPresentations, pres);
 }
 
+#ifdef LY_FIND_LEAKS
 /*
 **	Purpose:	Free the presentation list.
 **	Arguments:	void
@@ -195,6 +201,7 @@ PRIVATE void HTFreePresentations NOARGS
     HTList_delete(HTPresentations);
     HTPresentations = NULL;
 }
+#endif /* LY_FIND_LEAKS */
 
 /*	File buffering
 **	--------------
@@ -223,7 +230,7 @@ PUBLIC void HTInitInput ARGS1 (int,file_number)
 }
 
 PUBLIC int interrupted_in_htgetcharacter = 0;
-PUBLIC char HTGetCharacter NOARGS
+PUBLIC int HTGetCharacter NOARGS
 {
     char ch;
     interrupted_in_htgetcharacter = 0;
@@ -233,18 +240,15 @@ PUBLIC char HTGetCharacter NOARGS
 				 input_buffer, INPUT_BUFFER_SIZE);
 	    if (status <= 0) {
 		if (status == 0)
-		    return (char)EOF;
+		    return EOF;
 		if (status == HT_INTERRUPTED) {
-		    if (TRACE)
-			fprintf(stderr,
-				"HTFormat: Interrupted in HTGetCharacter\n");
+		    CTRACE((tfp, "HTFormat: Interrupted in HTGetCharacter\n"));
 		    interrupted_in_htgetcharacter = 1;
-		    return (char)EOF;
+		    return EOF;
 		}
-		if (TRACE)
-		    fprintf(stderr, "HTFormat: File read error %d\n", status);
-		return (char)EOF; /* -1 is returned by UCX
-				     at end of HTTP link */
+		CTRACE((tfp, "HTFormat: File read error %d\n", status));
+		return EOF; /* -1 is returned by UCX
+			       at end of HTTP link */
 	    }
 	    input_pointer = input_buffer;
 	    input_limit = input_buffer + status;
@@ -252,7 +256,7 @@ PUBLIC char HTGetCharacter NOARGS
 	ch = *input_pointer++;
     } while (ch == (char) 13); /* Ignore ASCII carriage return */
 
-    return FROMASCII(ch);
+    return FROMASCII((unsigned char)ch);
 }
 
 /*  Match maintype to any MIME type starting with maintype,
@@ -266,9 +270,8 @@ PRIVATE int half_match ARGS2(char *,trial_type, char *,target)
     if (!cp || *(cp+1) != '*')
 	return 0;
 
-    if (TRACE)
-	fprintf(stderr,"HTFormat: comparing %s and %s for half match\n",
-						      trial_type, target);
+    CTRACE((tfp, "HTFormat: comparing %s and %s for half match\n",
+		trial_type, target));
 
 	/* main type matches */
     if (!strncmp(trial_type, target, (cp-trial_type)-1))
@@ -277,6 +280,8 @@ PRIVATE int half_match ARGS2(char *,trial_type, char *,target)
     return 0;
 }
 
+#define WWW_WILDCARD_REP_OUT HTAtom_for("*")
+
 /*		Look up a presentation
 **		----------------------
 **
@@ -293,12 +298,10 @@ PRIVATE HTPresentation * HTFindPresentation ARGS3(
 	HTFormat,		rep_out,
 	HTPresentation*,	fill_in)
 {
-    HTAtom * wildcard = HTAtom_for("*");
+    HTAtom * wildcard = NULL; /* = HTAtom_for("*"); lookup when needed - kw */
 
-    if (TRACE)
-	fprintf(stderr,
-		"HTFormat: Looking up presentation for %s to %s\n",
-		HTAtom_name(rep_in), HTAtom_name(rep_out));
+    CTRACE((tfp, "HTFormat: Looking up presentation for %s to %s\n",
+		HTAtom_name(rep_in), HTAtom_name(rep_out)));
 
     /* don't do anymore do it in the Lynx code at startup LJM */
     /* if (!HTPresentations) HTFormatInit(); */ /* set up the list */
@@ -316,22 +319,21 @@ PRIVATE HTPresentation * HTFindPresentation ARGS3(
 	    pres = (HTPresentation *)HTList_objectAt(HTPresentations, i);
 	    if (pres->rep == rep_in) {
 		if (pres->rep_out == rep_out) {
-		    if (TRACE)
-			fprintf(stderr,
-				"FindPresentation: found exact match: %s\n",
-				HTAtom_name(pres->rep));
+		    CTRACE((tfp, "FindPresentation: found exact match: %s\n",
+				HTAtom_name(pres->rep)));
 		    return pres;
 
 		} else if (!fill_in) {
 		    continue;
-		} else if (pres->rep_out == wildcard) {
-		    if (!strong_wildcard_match)
-			strong_wildcard_match = pres;
-		    /* otherwise use the first one */
-		    if (TRACE)
-			fprintf(stderr,
-			     "StreamStack: found strong wildcard match: %s\n",
-				HTAtom_name(pres->rep));
+		} else {
+		    if (!wildcard) wildcard = WWW_WILDCARD_REP_OUT;
+		    if (pres->rep_out == wildcard) {
+			if (!strong_wildcard_match)
+			    strong_wildcard_match = pres;
+			/* otherwise use the first one */
+			CTRACE((tfp, "StreamStack: found strong wildcard match: %s\n",
+				    HTAtom_name(pres->rep)));
+		    }
 		}
 
 	    } else if (!fill_in) {
@@ -343,10 +345,8 @@ PRIVATE HTPresentation * HTFindPresentation ARGS3(
 		    if (!strong_subtype_wildcard_match)
 			strong_subtype_wildcard_match = pres;
 		    /* otherwise use the first one */
-		    if (TRACE)
-			fprintf(stderr,
-		     "StreamStack: found strong subtype wildcard match: %s\n",
-				HTAtom_name(pres->rep));
+		    CTRACE((tfp, "StreamStack: found strong subtype wildcard match: %s\n",
+				HTAtom_name(pres->rep)));
 		}
 	    }
 
@@ -355,13 +355,12 @@ PRIVATE HTPresentation * HTFindPresentation ARGS3(
 		    if (!weak_wildcard_match)
 			weak_wildcard_match = pres;
 		    /* otherwise use the first one */
-		    if (TRACE)
-			fprintf(stderr,
-			    "StreamStack: found weak wildcard match: %s\n",
-				HTAtom_name(pres->rep_out));
-		}
-		if (pres->rep_out == wildcard) {
-		    if (!last_default_match)
+		    CTRACE((tfp, "StreamStack: found weak wildcard match: %s\n",
+				HTAtom_name(pres->rep_out)));
+
+		} else if (!last_default_match) {
+		    if (!wildcard) wildcard = WWW_WILDCARD_REP_OUT;
+		    if (pres->rep_out == wildcard)
 			 last_default_match = pres;
 		    /* otherwise use the first one */
 		}
@@ -402,36 +401,44 @@ PUBLIC HTStream * HTStreamStack ARGS4(
 {
     HTPresentation temp;
     HTPresentation *match;
+    HTStream *result;
 
-    if (TRACE)
-	fprintf(stderr,
-		"HTFormat: Constructing stream stack for %s to %s\n",
-		HTAtom_name(rep_in), HTAtom_name(rep_out));
+    CTRACE((tfp, "HTFormat: Constructing stream stack for %s to %s\n",
+		HTAtom_name(rep_in), HTAtom_name(rep_out)));
 
     /* don't return on WWW_SOURCE some people might like
      * to make use of the source!!!!  LJM
-     *//*
+     */
+#if 0
     if (rep_out == WWW_SOURCE || rep_out == rep_in)
-	return sink;  LJM */
+	return sink;	/*  LJM */
+#endif
 
-    if (rep_out == rep_in)
-	return sink;
+    if (rep_out == rep_in) {
+	result = sink;
 
-    if ((match = HTFindPresentation(rep_in, rep_out, &temp))) {
+    } else if ((match = HTFindPresentation(rep_in, rep_out, &temp))) {
 	if (match == &temp) {
-	    if (TRACE)
-		fprintf(stderr,
-			"StreamStack: Using %s\n", HTAtom_name(temp.rep_out));
+	    CTRACE((tfp, "StreamStack: Using %s\n", HTAtom_name(temp.rep_out)));
 	} else {
-	    if (TRACE)
-		fprintf(stderr,
-			"StreamStack: found exact match: %s\n",
-			HTAtom_name(match->rep));
+	    CTRACE((tfp, "StreamStack: found exact match: %s\n",
+			HTAtom_name(match->rep)));
 	}
-	return (*match->converter)(match, anchor, sink);
+	result = (*match->converter)(match, anchor, sink);
     } else {
-	return NULL;
+	result = NULL;
     }
+    if (TRACE) {
+	if (result && result->isa && result->isa->name) {
+	    CTRACE((tfp, "StreamStack: Returning \"%s\"\n", result->isa->name));
+	} else if (result) {
+	    CTRACE((tfp, "StreamStack: Returning *unknown* stream!\n"));
+	} else {
+	    CTRACE((tfp, "StreamStack: Returning NULL!\n"));
+	    CTRACE_FLUSH(tfp);	/* a crash may be imminent... - kw */
+	}
+    }
+    return result;
 }
 
 /*		Put a presentation near start of list
@@ -464,12 +471,10 @@ PUBLIC float HTStackValue ARGS4(
 	float,			initial_value,
 	long int,		length)
 {
-    HTAtom * wildcard = HTAtom_for("*");
+    HTAtom * wildcard = WWW_WILDCARD_REP_OUT;
 
-    if (TRACE)
-	fprintf(stderr,
-		"HTFormat: Evaluating stream stack for %s worth %.3f to %s\n",
-		HTAtom_name(rep_in), initial_value, HTAtom_name(rep_out));
+    CTRACE((tfp, "HTFormat: Evaluating stream stack for %s worth %.3f to %s\n",
+		HTAtom_name(rep_in), initial_value, HTAtom_name(rep_out)));
 
     if (rep_out == WWW_SOURCE || rep_out == rep_in)
 	return 0.0;
@@ -494,20 +499,101 @@ PUBLIC float HTStackValue ARGS4(
 	}
     }
 
-    return -1e30;		/* Really bad */
+    return (float) -1e30;	/* Really bad */
 
 }
 
+/*	Display the page while transfer in progress
+**	-------------------------------------------
+**
+**   Repaint the page only when necessary.
+**   This is a traverse call for HText_pageDisplay() - it works!.
+**
+*/
+PUBLIC void HTDisplayPartial NOARGS
+{
+#ifdef DISP_PARTIAL
+    if (display_partial) {
+	/*
+	**  HText_getNumOfLines() = "current" number of complete lines received
+	**  NumOfLines_partial = number of lines at the moment of last repaint.
+	**  (we update NumOfLines_partial only when we repaint the display.)
+	**
+	**  display_partial could only be enabled in HText_new()
+	**  so a new HTMainText object available - all HText_ functions use it,
+	**  lines counter HText_getNumOfLines() in particular.
+	**
+	**  Otherwise HTMainText holds info from the previous document
+	**  and we may repaint it instead of the new one:
+	**  prev doc scrolled to the first line (=Newline_partial)
+	**  is not good looking :-)	  23 Aug 1998 Leonid Pauzner
+	**
+	**  So repaint the page only when necessary:
+	*/
+	int Newline_partial = LYGetNewline();
+
+	if (((Newline_partial + display_lines) - 1 > NumOfLines_partial)
+		/* current page not complete... */
+	&& (partial_threshold > 0 ?
+		((Newline_partial + partial_threshold) -1 <= HText_getNumOfLines()) :
+		((Newline_partial + display_lines) - 1 <= HText_getNumOfLines()))
+		/*
+		 * Originally we rendered by increments of 2 lines,
+		 * but that got annoying on slow network connections.
+		 * Then we switched to full-pages.  Now it's configurable.
+		 * If partial_threshold <= 0, then it's a full page
+		 */
+	) {
+	    NumOfLines_partial = HText_getNumOfLines();
+	    LYMainLoop_pageDisplay(Newline_partial);
+	}
+    }
+#else /* nothing */
+#endif  /* DISP_PARTIAL */
+}
+
+/* Put this as early as possible, OK just after HTDisplayPartial() */
+PUBLIC void HTFinishDisplayPartial NOARGS
+{
+#ifdef DISP_PARTIAL
+		    /*
+		     *  End of incremental rendering stage here.
+		     */
+		    display_partial = FALSE;
+#endif /* DISP_PARTIAL */
+}
+
 /*	Push data from a socket down a stream
 **	-------------------------------------
 **
 **   This routine is responsible for creating and PRESENTING any
 **   graphic (or other) objects described by the file.
 **
-**   The file number given is assumed to be a TELNET stream ie containing
+**   The file number given is assumed to be a TELNET stream, i.e., containing
 **   CRLF at the end of lines which need to be stripped to LF for unix
 **   when the format is textual.
 **
+**  State of socket and target stream on entry:
+**			socket (file_number) assumed open,
+**			target (sink) assumed valid.
+**
+**  Return values:
+**	HT_INTERRUPTED  Interruption or error after some data received.
+**	-2		Unexpected disconnect before any data received.
+**	-1		Interruption or error before any data received, or
+**			(UNIX) other read error before any data received, or
+**			download cancelled.
+**	HT_LOADED	Normal close of socket (end of file indication
+**			received), or
+**			unexpected disconnect after some data received, or
+**			other read error after some data received, or
+**			(not UNIX) other read error before any data received.
+**
+**  State of socket and target stream on return depends on return value:
+**	HT_INTERRUPTED	socket still open, target aborted.
+**	-2		socket still open, target stream still valid.
+**	-1		socket still open, target aborted.
+**	otherwise	socket closed,	target stream still valid.
 */
 PUBLIC int HTCopy ARGS4(
 	HTParentAnchor *,	anchor,
@@ -516,9 +602,15 @@ PUBLIC int HTCopy ARGS4(
 	HTStream*,		sink)
 {
     HTStreamClass targetClass;
-    char line[256];
-    int bytes = 0;
+    BOOL suppress_readprogress = NO;
+    int bytes;
     int rv = 0;
+#ifdef _WINDOWS	/* 1997/11/11 (Tue) 15:18:16 */
+    long file_length;
+    extern int bytes_already_read;
+
+    file_length = anchor->content_length;
+#endif
 
     /*	Push the data down the stream
     */
@@ -528,6 +620,7 @@ PUBLIC int HTCopy ARGS4(
     **
     **	This operation could be put into a main event loop
     */
+    HTReadProgress(bytes = 0, 0);
     for (;;) {
 	int status;
 
@@ -539,7 +632,7 @@ PUBLIC int HTCopy ARGS4(
 	}
 
 	if (HTCheckForInterrupt()) {
-	    _HTProgress ("Data transfer interrupted.");
+	    _HTProgress (TRANSFER_INTERRUPTED);
 	    (*targetClass._abort)(sink, NULL);
 	    if (bytes)
 		rv = HT_INTERRUPTED;
@@ -554,7 +647,7 @@ PUBLIC int HTCopy ARGS4(
 	    if (status == 0) {
 		break;
 	    } else if (status == HT_INTERRUPTED) {
-		_HTProgress ("Data transfer interrupted.");
+		_HTProgress (TRANSFER_INTERRUPTED);
 		(*targetClass._abort)(sink, NULL);
 		if (bytes)
 		    rv = HT_INTERRUPTED;
@@ -562,10 +655,13 @@ PUBLIC int HTCopy ARGS4(
 		    rv = -1;
 		goto finished;
 	    } else if (SOCKET_ERRNO == ENOTCONN ||
+#ifdef _WINDOWS	/* 1997/11/10 (Mon) 16:57:18 */
+		       SOCKET_ERRNO == ETIMEDOUT ||
+#endif
 		       SOCKET_ERRNO == ECONNRESET ||
 		       SOCKET_ERRNO == EPIPE) {
 		/*
-		 *  Arrrrgh, HTTP 0/1 compability problem, maybe.
+		 *  Arrrrgh, HTTP 0/1 compatibility problem, maybe.
 		 */
 		if (bytes <= 0) {
 		    /*
@@ -575,20 +671,65 @@ PUBLIC int HTCopy ARGS4(
 		    rv = -2;
 		    goto finished;
 		} else {
+#ifdef UNIX
+		   /*
+		    *  Treat what we've received already as the complete
+		    *  transmission, but not without giving the user
+		    *  an alert.  I don't know about all the different
+		    *  TCP stacks for VMS etc., so this is currently
+		    *  only for UNIX. - kw
+		    */
+		   HTInetStatus("NETREAD");
+		   HTAlert("Unexpected server disconnect.");
+		   CTRACE((tfp,
+	    "HTCopy: Unexpected server disconnect. Treating as completed.\n"));
+		   status = 0;
+		   break;
+#else  /* !UNIX */
 		   /*
 		    *  Treat what we've gotten already
 		    *  as the complete transmission. - FM
 		    */
-		   if (TRACE)
-		       fprintf(stderr,
-	    "HTCopy: Unexpected server disconnect. Treating as completed.\n");
+		   CTRACE((tfp,
+	    "HTCopy: Unexpected server disconnect.  Treating as completed.\n"));
 		   status = 0;
 		   break;
+#endif /* UNIX */
 		}
+#ifdef UNIX
+	    } else {		/* status < 0 and other errno */
+		/*
+		 *  Treat what we've received already as the complete
+		 *  transmission, but not without giving the user
+		 *  an alert.  I don't know about all the different
+		 *  TCP stacks for VMS etc., so this is currently
+		 *  only for UNIX. - kw
+		 */
+		HTInetStatus("NETREAD");
+		HTAlert("Unexpected read error.");
+		if (bytes) {
+		    (void)NETCLOSE(file_number);
+		    rv = HT_LOADED;
+		} else {
+		    (*targetClass._abort)(sink, NULL);
+		    rv = -1;
+		}
+		goto finished;
+#endif
 	    }
 	    break;
 	}
 
+	/*
+	 *  Suppress ReadProgress messages when collecting a redirection
+	 *  message, at least initially (unless/until anchor->content_type
+	 *  gets changed, probably by the MIME message parser).  That way
+	 *  messages put up by the HTTP module or elsewhere can linger in
+	 *  the statusline for a while. - kw
+	 */
+	suppress_readprogress = (anchor && anchor->content_type &&
+				 !strcmp(anchor->content_type,
+					 "message/x-http-redirection"));
 #ifdef NOT_ASCII
 	{
 	    char * p;
@@ -599,22 +740,19 @@ PUBLIC int HTCopy ARGS4(
 #endif /* NOT_ASCII */
 
 	(*targetClass.put_block)(sink, input_buffer, status);
-
 	bytes += status;
-	if (anchor && anchor->content_length > 0)
-	    sprintf(line, "Read %d of %d bytes of data.",
-			  bytes, anchor->content_length);
-	else
-	    sprintf(line, "Read %d bytes of data.", bytes);
-	HTProgress(line);
+	if (!suppress_readprogress)
+	    HTReadProgress(bytes, anchor ? anchor->content_length : 0);
+	HTDisplayPartial();
 
     } /* next bufferload */
 
-    _HTProgress("Data transfer complete");
+    _HTProgress(TRANSFER_COMPLETE);
     (void)NETCLOSE(file_number);
     rv = HT_LOADED;
 
 finished:
+    HTFinishDisplayPartial();
     return(rv);
 }
 
@@ -625,14 +763,25 @@ finished:
 **   graphic (or other) objects described by the file.
 **
 **
+**  State of file and target stream on entry:
+**			FILE* (fp) assumed open,
+**			target (sink) assumed valid.
+**
+**  Return values:
+**	HT_INTERRUPTED  Interruption after some data read.
+**	HT_PARTIAL_CONTENT	Error after some data read.
+**	-1		Error before any data read.
+**	HT_LOADED	Normal end of file indication on reading.
+**
+**  State of file and target stream on return:
+**	always		fp still open, target stream still valid.
 */
 PUBLIC int HTFileCopy ARGS2(
-	FILE *, 		fp,
+	FILE *,			fp,
 	HTStream*,		sink)
 {
     HTStreamClass targetClass;
-    char line[256];
-    int status, bytes = 0, nreads = 0, nprogr = 0;
+    int status, bytes;
     int rv = HT_OK;
 
     /*	Push the data down the stream
@@ -641,18 +790,16 @@ PUBLIC int HTFileCopy ARGS2(
 
     /*	Push binary from socket down sink
     */
+    HTReadProgress(bytes = 0, 0);
     for (;;) {
 	status = fread(input_buffer, 1, INPUT_BUFFER_SIZE, fp);
-	nreads++;
 	if (status == 0) { /* EOF or error */
 	    if (ferror(fp) == 0) {
 		rv = HT_LOADED;
 		break;
 	    }
-	    if (TRACE)
-		fprintf(stderr,
-			"HTFormat: Read error, read returns %d\n",
-			ferror(fp));
+	    CTRACE((tfp, "HTFormat: Read error, read returns %d\n",
+			ferror(fp)));
 	    if (bytes) {
 		rv = HT_PARTIAL_CONTENT;
 	    } else {
@@ -660,47 +807,92 @@ PUBLIC int HTFileCopy ARGS2(
 	    }
 	    break;
 	}
-	(*targetClass.put_block)(sink, input_buffer, status);
 
+	(*targetClass.put_block)(sink, input_buffer, status);
 	bytes += status;
-	if (nreads >= 100) {
-	    /*
-	    **	Show progress messages for local files, and check for
-	    **	user interruption.  Start doing so only after a certain
-	    **	number of reads have been done, and don't update it on
-	    **	every read (normally reading in a local file should be
-	    **	speedy). - KW
-	    */
-	    if (nprogr == 0) {
-		if (bytes < 1024000) {
-		    sprintf(line, "Read %d bytes of data.", bytes);
-		} else {
-		    sprintf(line, "Read %d KB of data. %s",
-				  bytes/1024,
-		    "(Press 'z' if you want to abort loading.)");
-		}
-		HTProgress(line);
-		if (HTCheckForInterrupt()) {
-		    _HTProgress ("Data transfer interrupted.");
-		    if (bytes) {
-			rv = HT_INTERRUPTED;
-		    } else {
-			rv = -1;
-		    }
-		    break;
-		}
-		nprogr++;
-	    } else if (nprogr == 25) {
-		nprogr = 0;
+	HTReadProgress(bytes, 0);
+	/*  Suppress last screen update in partial mode - a regular update
+	 *  under control of mainloop() should follow anyway. - kw
+	 */
+#ifdef DISP_PARTIAL
+	if (display_partial && bytes != HTMainAnchor->content_length)
+	    HTDisplayPartial();
+#endif
+
+	if (HTCheckForInterrupt()) {
+	    _HTProgress (TRANSFER_INTERRUPTED);
+	    if (bytes) {
+		rv = HT_INTERRUPTED;
 	    } else {
-		nprogr++;
+		rv = -1;
 	    }
+	    break;
 	}
     } /* next bufferload */
 
+    HTFinishDisplayPartial();
     return rv;
 }
 
+#ifdef SOURCE_CACHE
+/*	Push data from an HTChunk down a stream
+**	---------------------------------------
+**
+**   This routine is responsible for creating and PRESENTING any
+**   graphic (or other) objects described by the file.
+**
+**  State of memory and target stream on entry:
+**			HTChunk* (chunk) and target (sink) assumed valid.
+**
+**  Return values:
+**	HT_LOADED	All data sent.
+**	HT_INTERRUPTED  Interruption after some data read.
+**
+**  State of memory and target stream on return:
+**	always		chunk unchanged, target stream still valid.
+*/
+PUBLIC int HTMemCopy ARGS2(
+	HTChunk *,		chunk,
+	HTStream *,		sink)
+{
+    HTStreamClass targetClass;
+    int bytes = 0;
+    CONST char *data = chunk->data;
+    int rv = HT_OK;
+
+    targetClass = *(sink->isa);
+    HTReadProgress(0, 0);
+    for (;;) {
+	/* Push the data down the stream a piece at a time, in case we're
+	** running a large document on a slow machine.
+	*/
+	int n = INPUT_BUFFER_SIZE;
+	if (n > chunk->size - bytes)
+	    n = chunk->size - bytes;
+	if (n == 0)
+	    break;
+	(*targetClass.put_block)(sink, data, n);
+	bytes += n;
+	data += n;
+	HTReadProgress(bytes, 0);
+	HTDisplayPartial();
+
+	if (HTCheckForInterrupt()) {
+	    _HTProgress (TRANSFER_INTERRUPTED);
+	    if (bytes) {
+		rv = HT_INTERRUPTED;
+	    } else {
+		rv = -1;
+	    }
+	    break;
+	}
+    }
+
+    HTFinishDisplayPartial();
+    return rv;
+}
+#endif
+
 #ifdef USE_ZLIB
 /*	Push data from a gzip file pointer down a stream
 **	-------------------------------------
@@ -709,14 +901,25 @@ PUBLIC int HTFileCopy ARGS2(
 **   graphic (or other) objects described by the file.
 **
 **
+**  State of file and target stream on entry:
+**		      gzFile (gzfp) assumed open (should have gzipped content),
+**		      target (sink) assumed valid.
+**
+**  Return values:
+**	HT_INTERRUPTED  Interruption after some data read.
+**	HT_PARTIAL_CONTENT	Error after some data read.
+**	-1		Error before any data read.
+**	HT_LOADED	Normal end of file indication on reading.
+**
+**  State of file and target stream on return:
+**	always		gzfp still open, target stream still valid.
 */
 PRIVATE int HTGzFileCopy ARGS2(
-	gzFile, 		gzfp,
+	gzFile,			gzfp,
 	HTStream*,		sink)
 {
     HTStreamClass targetClass;
-    char line[256];
-    int status, bytes = 0, nreads = 0, nprogr = 0;
+    int status, bytes;
     int gzerrnum;
     int rv = HT_OK;
 
@@ -724,23 +927,21 @@ PRIVATE int HTGzFileCopy ARGS2(
     */
     targetClass = *(sink->isa); /* Copy pointers to procedures */
 
-    /*	read and inflate gzipped file, and push binary down sink
+    /*	read and inflate gzip'd file, and push binary down sink
     */
+    HTReadProgress(bytes = 0, 0);
     for (;;) {
 	status = gzread(gzfp, input_buffer, INPUT_BUFFER_SIZE);
-	nreads++;
 	if (status <= 0) { /* EOF or error */
 	    if (status == 0) {
 		rv = HT_LOADED;
 		break;
 	    }
+	    CTRACE((tfp, "HTGzFileCopy: Read error, gzread returns %d\n",
+			status));
+	    CTRACE((tfp, "gzerror   : %s\n",
+			gzerror(gzfp, &gzerrnum)));
 	    if (TRACE) {
-		fprintf(stderr,
-			"HTGzFileCopy: Read error, gzread returns %d\n",
-			status);
-		fprintf(stderr,
-			"gzerror   : %s\n",
-			gzerror(gzfp, &gzerrnum));
 		if (gzerrnum == Z_ERRNO)
 		    perror("gzerror   ");
 	    }
@@ -751,45 +952,24 @@ PRIVATE int HTGzFileCopy ARGS2(
 	    }
 	    break;
 	}
-	(*targetClass.put_block)(sink, input_buffer, status);
 
+	(*targetClass.put_block)(sink, input_buffer, status);
 	bytes += status;
-	if (nreads >= 100) {
-	    /*
-	    **	Show progress messages for local files, and check for
-	    **	user interruption.  Start doing so only after a certain
-	    **	number of reads have been done, and don't update it on
-	    **	every read (normally reading in a local file should be
-	    **	speedy). - KW
-	    */
-	    if (nprogr == 0) {
-		if (bytes < 1024000) {
-		    sprintf(line,
-			    "Read %d uncompressed bytes of data.", bytes);
-		} else {
-		    sprintf(line, "Read %d uncompressed KB of data. %s",
-				  bytes/1024,
-		    "(Press 'z' to abort.)");
-		}
-		HTProgress(line);
-		if (HTCheckForInterrupt()) {
-		    _HTProgress ("Data transfer interrupted.");
-		    if (bytes) {
-			rv = HT_INTERRUPTED;
-		    } else {
-			rv = -1;
-		    }
-		    break;
-		}
-		nprogr++;
-	    } else if (nprogr == 25) {
-		nprogr = 0;
+	HTReadProgress(bytes, -1);
+	HTDisplayPartial();
+
+	if (HTCheckForInterrupt()) {
+	    _HTProgress (TRANSFER_INTERRUPTED);
+	    if (bytes) {
+		rv = HT_INTERRUPTED;
 	    } else {
-		nprogr++;
+		rv = -1;
 	    }
+	    break;
 	}
     } /* next bufferload */
 
+    HTFinishDisplayPartial();
     return rv;
 }
 #endif /* USE_ZLIB */
@@ -811,7 +991,7 @@ PUBLIC void HTCopyNoCR ARGS3(
 	HTStream*,		sink)
 {
     HTStreamClass targetClass;
-    char character;
+    int character;
 
     /*	Push the data, ignoring CRLF, down the stream
     */
@@ -825,9 +1005,9 @@ PUBLIC void HTCopyNoCR ARGS3(
     HTInitInput(file_number);
     for (;;) {
 	character = HTGetCharacter();
-	if (character == (char)EOF)
+	if (character == EOF)
 	    break;
-	(*targetClass.put_character)(sink, character);
+	(*targetClass.put_character)(sink, (unsigned char) character);
     }
 }
 
@@ -840,6 +1020,30 @@ PUBLIC void HTCopyNoCR ARGS3(
 **   CRLF at the end of lines which need to be stripped to LF for unix
 **   when the format is textual.
 **
+**  State of socket and target stream on entry:
+**			socket (file_number) assumed open,
+**			target (sink) usually NULL (will call stream stack).
+**
+**  Return values:
+**	HT_INTERRUPTED  Interruption or error after some data received.
+**	-501		Stream stack failed (cannot present or convert).
+**	-2		Unexpected disconnect before any data received.
+**	-1		Stream stack failed (cannot present or convert), or
+**			Interruption or error before any data received, or
+**			(UNIX) other read error before any data received, or
+**			download cancelled.
+**	HT_LOADED	Normal close of socket (end of file indication
+**			received), or
+**			unexpected disconnect after some data received, or
+**			other read error after some data received, or
+**			(not UNIX) other read error before any data received.
+**
+**  State of socket and target stream on return depends on return value:
+**	HT_INTERRUPTED	socket still open, target aborted.
+**	-501		socket still open, target stream NULL.
+**	-2		socket still open, target freed.
+**	-1		socket still open, target stream aborted or NULL.
+**	otherwise	socket closed,	target stream freed.
 */
 PUBLIC int HTParseSocket ARGS5(
 	HTFormat,		rep_in,
@@ -855,27 +1059,27 @@ PUBLIC int HTParseSocket ARGS5(
     stream = HTStreamStack(rep_in, format_out, sink, anchor);
 
     if (!stream) {
-	char buffer[1024];	/* @@@@@@@@ */
+	char *buffer = 0;
 	if (LYCancelDownload) {
 	    LYCancelDownload = FALSE;
 	    return -1;
 	}
-	sprintf(buffer, "Sorry, can't convert from %s to %s.",
+	HTSprintf0(&buffer, CANNOT_CONVERT_I_TO_O,
 		HTAtom_name(rep_in), HTAtom_name(format_out));
-	if (TRACE)
-	    fprintf(stderr, "HTFormat: %s\n", buffer);
-	return HTLoadError(sink, 501, buffer); /* returns -501 */
+	CTRACE((tfp, "HTFormat: %s\n", buffer));
+	rv = HTLoadError(sink, 501, buffer); /* returns -501 */
+	FREE(buffer);
+    } else {
+	/*
+	** Push the data, don't worry about CRLF we can strip them later.
+	*/
+	targetClass = *(stream->isa);	/* Copy pointers to procedures */
+	rv = HTCopy(anchor, file_number, NULL, stream);
+	if (rv != -1 && rv != HT_INTERRUPTED)
+	    (*targetClass._free)(stream);
     }
-
-    /*
-    ** Push the data, don't worry about CRLF we can strip them later.
-    */
-    targetClass = *(stream->isa);	/* Copy pointers to procedures */
-    rv = HTCopy(anchor, file_number, NULL, stream);
-    if (rv != -1 && rv != HT_INTERRUPTED)
-	(*targetClass._free)(stream);
-
-    return rv; /* full: HT_LOADED;  partial: HT_INTERRUPTED;  no bytes: -1 */
+    return rv;
+    /* Originally:  full: HT_LOADED;  partial: HT_INTERRUPTED;  no bytes: -1 */
 }
 
 /*	Parse a file given format and file pointer
@@ -887,33 +1091,50 @@ PUBLIC int HTParseSocket ARGS5(
 **   CRLF at the end of lines which need to be stripped to \n for unix
 **   when the format is textual.
 **
+**  State of file and target stream on entry:
+**			FILE* (fp) assumed open,
+**			target (sink) usually NULL (will call stream stack).
+**
+**  Return values:
+**	-501		Stream stack failed (cannot present or convert).
+**	-1		Download cancelled.
+**	HT_NO_DATA	Error before any data read.
+**	HT_PARTIAL_CONTENT	Interruption or error after some data read.
+**	HT_LOADED	Normal end of file indication on reading.
+**
+**  State of file and target stream on return:
+**	always		fp still open; target freed, aborted, or NULL.
 */
 PUBLIC int HTParseFile ARGS5(
 	HTFormat,		rep_in,
 	HTFormat,		format_out,
 	HTParentAnchor *,	anchor,
-	FILE *, 		fp,
+	FILE *,			fp,
 	HTStream*,		sink)
 {
     HTStream * stream;
     HTStreamClass targetClass;
     int rv;
 
-    stream = HTStreamStack(rep_in,
-			format_out,
-			sink , anchor);
+#ifdef SH_EX		/* 1998/01/04 (Sun) 16:04:09 */
+    if (fp == NULL)
+	return HT_LOADED;
+#endif
+
+    stream = HTStreamStack(rep_in, format_out, sink, anchor);
 
     if (!stream) {
-	char buffer[1024];	/* @@@@@@@@ */
+	char *buffer = 0;
 	if (LYCancelDownload) {
 	    LYCancelDownload = FALSE;
 	    return -1;
 	}
-	sprintf(buffer, "Sorry, can't convert from %s to %s.",
+	HTSprintf0(&buffer, CANNOT_CONVERT_I_TO_O,
 		HTAtom_name(rep_in), HTAtom_name(format_out));
-	if (TRACE)
-	    fprintf(stderr, "HTFormat(in HTParseFile): %s\n", buffer);
-	return HTLoadError(sink, 501, buffer);
+	CTRACE((tfp, "HTFormat(in HTParseFile): %s\n", buffer));
+	rv = HTLoadError(sink, 501, buffer);
+	FREE(buffer);
+	return rv;
     }
 
     /*	Push the data down the stream
@@ -938,9 +1159,57 @@ PUBLIC int HTParseFile ARGS5(
 	return HT_LOADED;
 }
 
+#ifdef SOURCE_CACHE
+/*	Parse a document in memory given format and memory block pointer
+**
+**   This routine is responsible for creating and PRESENTING any
+**   graphic (or other) objects described by the file.
+**
+**  State of memory and target stream on entry:
+**			HTChunk* (chunk) assumed valid,
+**			target (sink) usually NULL (will call stream stack).
+**
+**  Return values:
+**	-501		Stream stack failed (cannot present or convert).
+**	HT_LOADED	All data sent.
+**
+**  State of memory and target stream on return:
+**	always		chunk unchanged; target freed, aborted, or NULL.
+*/
+PUBLIC int HTParseMem ARGS5(
+	HTFormat,		rep_in,
+	HTFormat,		format_out,
+	HTParentAnchor *,	anchor,
+	HTChunk *,		chunk,
+	HTStream *,		sink)
+{
+    HTStream * stream;
+    HTStreamClass targetClass;
+    int rv;
+
+    stream = HTStreamStack(rep_in, format_out, sink, anchor);
+    if (!stream) {
+	char *buffer = 0;
+	HTSprintf0(&buffer, CANNOT_CONVERT_I_TO_O,
+		   HTAtom_name(rep_in), HTAtom_name(format_out));
+	CTRACE((tfp, "HTFormat(in HTParseMem): %s\n", buffer));
+	rv = HTLoadError(sink, 501, buffer);
+	FREE(buffer);
+	return rv;
+    }
+
+    /* Push the data down the stream
+    */
+    targetClass = *(stream->isa);
+    rv = HTMemCopy(chunk, stream);
+    (*targetClass._free)(stream);
+    return HT_LOADED;
+}
+#endif
+
 #ifdef USE_ZLIB
 PRIVATE int HTCloseGzFile ARGS1(
-	gzFile, 		gzfp)
+	gzFile,			gzfp)
 {
     int gzres;
     if (gzfp == NULL)
@@ -950,40 +1219,54 @@ PRIVATE int HTCloseGzFile ARGS1(
 	if (gzres == Z_ERRNO) {
 	    perror("gzclose   ");
 	} else if (gzres != Z_OK) {
-	    fprintf(stderr, "gzclose   : error number %d\n", gzres);
+	    CTRACE((tfp, "gzclose   : error number %d\n", gzres));
 	}
     }
     return(gzres);
 }
 
+/*	HTParseGzFile
+**
+**  State of file and target stream on entry:
+**			gzFile (gzfp) assumed open,
+**			target (sink) usually NULL (will call stream stack).
+**
+**  Return values:
+**	-501		Stream stack failed (cannot present or convert).
+**	-1		Download cancelled.
+**	HT_NO_DATA	Error before any data read.
+**	HT_PARTIAL_CONTENT	Interruption or error after some data read.
+**	HT_LOADED	Normal end of file indication on reading.
+**
+**  State of file and target stream on return:
+**	always		gzfp closed; target freed, aborted, or NULL.
+*/
 PUBLIC int HTParseGzFile ARGS5(
 	HTFormat,		rep_in,
 	HTFormat,		format_out,
 	HTParentAnchor *,	anchor,
-	gzFile, 		gzfp,
+	gzFile,			gzfp,
 	HTStream*,		sink)
 {
     HTStream * stream;
     HTStreamClass targetClass;
     int rv;
 
-    stream = HTStreamStack(rep_in,
-			format_out,
-			sink , anchor);
+    stream = HTStreamStack(rep_in, format_out, sink, anchor);
 
     if (!stream) {
-	char buffer[1024];	/* @@@@@@@@ */
-	extern char LYCancelDownload;
+	char *buffer = 0;
 	HTCloseGzFile(gzfp);
 	if (LYCancelDownload) {
 	    LYCancelDownload = FALSE;
 	    return -1;
 	}
-	sprintf(buffer, "Sorry, can't convert from %s to %s.",
+	HTSprintf0(&buffer, CANNOT_CONVERT_I_TO_O,
 		HTAtom_name(rep_in), HTAtom_name(format_out));
-	if (TRACE)
-	    fprintf(stderr, "HTFormat(in HTParseGzFile): %s\n", buffer);
-	return HTLoadError(sink, 501, buffer);
+	CTRACE((tfp, "HTFormat(in HTParseGzFile): %s\n", buffer));
+	rv = HTLoadError(sink, 501, buffer);
+	FREE(buffer);
+	return rv;
     }
 
     /*	Push the data down the stream
@@ -1032,7 +1315,7 @@ PRIVATE void NetToText_put_character ARGS2(HTStream *, me, char, net_char)
 	    me->sink->isa->put_character(me->sink, CR); /* leftover */
 	}
     }
-    me->had_cr = (c == CR);
+    me->had_cr = (BOOL) (c == CR);
     if (!me->had_cr)
 	me->sink->isa->put_character(me->sink, c);		/* normal */
 }
@@ -1091,3 +1374,53 @@ PUBLIC HTStream * HTNetToText ARGS1(HTStream *, sink)
     return me;
 }
 
+PRIVATE HTStream	HTBaseStreamInstance;		      /* Made static */
+/*
+**	ERROR STREAM
+**	------------
+**	There is only one error stream shared by anyone who wants a
+**	generic error returned from all stream methods.
+*/
+PRIVATE void HTErrorStream_put_character ARGS2(HTStream *, me GCC_UNUSED, char, c GCC_UNUSED)
+{
+    LYCancelDownload = TRUE;
+}
+
+PRIVATE void HTErrorStream_put_string ARGS2(HTStream *, me GCC_UNUSED, CONST char *, s)
+{
+    if (s && *s)
+	LYCancelDownload = TRUE;
+}
+
+PRIVATE void HTErrorStream_write ARGS3(HTStream *, me GCC_UNUSED, CONST char *, s, int, l)
+{
+    if (l && s)
+	LYCancelDownload = TRUE;
+}
+
+PRIVATE void HTErrorStream_free ARGS1(HTStream *, me GCC_UNUSED)
+{
+    return;
+}
+
+PRIVATE void HTErrorStream_abort ARGS2(HTStream *, me GCC_UNUSED, HTError, e GCC_UNUSED)
+{
+    return;
+}
+
+PRIVATE CONST HTStreamClass HTErrorStreamClass =
+{
+    "ErrorStream",
+    HTErrorStream_free,
+    HTErrorStream_abort,
+    HTErrorStream_put_character,
+    HTErrorStream_put_string,
+    HTErrorStream_write
+};
+
+PUBLIC HTStream * HTErrorStream NOARGS
+{
+    CTRACE((tfp, "ErrorStream. Created\n"));
+    HTBaseStreamInstance.isa = &HTErrorStreamClass;    /* The rest is random */
+    return &HTBaseStreamInstance;
+}
