@@ -5,11 +5,14 @@
 #include <LYGlobalDefs.h>
 #include <LYJump.h>
 #include <LYKeymap.h>
-#include <LYSignal.h>
 #include <GridText.h>
 
 #include <LYLeaks.h>
 
+#ifdef _WINDOWS
+#include <stdlib.h>	/* bsearch() */
+#endif
+
 #ifdef VMS
 #include <fab.h>
 #endif /* VMS */
@@ -54,16 +57,14 @@ PUBLIC void LYJumpTable_free NOARGS
  */
 PUBLIC void LYAddJumpShortcut ARGS2(HTList *, historyp, char *,shortcut)
 {
-    char *new;
+    char *new = NULL;
     char *old;
     HTList *cur =  historyp;
 
     if (!historyp || !(shortcut && *shortcut))
 	return;
 
-    if ((new = (char *)calloc(1, (strlen(shortcut) + 1))) == NULL)
-	outofmem(__FILE__, "HTAddJumpShortcut");
-    strcpy(new, shortcut);
+    StrAllocCopy(new, shortcut);
 
     while (NULL != (old = (char *)HTList_nextObject(cur))) {
 	if (!strcmp(old, new)) {
@@ -85,10 +86,9 @@ PUBLIC BOOL LYJumpInit ARGS1 (char *, config)
     /*
      * Create a JumpTable structure.
      */
-    jtp = (struct JumpTable *) calloc(1, sizeof(*jtp));
+    jtp = typecalloc(struct JumpTable);
     if (jtp == NULL) {
-	perror("Out of memory in LYJumpInit");
-	return FALSE;
+	outofmem(__FILE__, "LYJumpInit");
     }
 
     /*
@@ -111,8 +111,10 @@ PUBLIC BOOL LYJumpInit ARGS1 (char *, config)
 	return FALSE;
     }
     StrAllocCopy(jtp->file, cp);
+#ifdef LY_FIND_LEAKS
     if (!JThead)
 	atexit(LYJumpTable_free);
+#endif /* LY_FIND_LEAKS */
 
     /*
      * Get the key, if present.
@@ -144,10 +146,9 @@ PUBLIC BOOL LYJumpInit ARGS1 (char *, config)
 	StrAllocCopy(JThead->msg, jumpprompt);
 	if (!jumpfile)
 	    StrAllocCopy(jumpfile, JThead->file);
-	jtp = (struct JumpTable *) calloc(1, sizeof(*jtp));
+	jtp = typecalloc(struct JumpTable);
 	if (jtp == NULL) {
-	    perror("Out of memory in LYJumpInit");
-	    return FALSE;
+	    outofmem(__FILE__, "LYJumpInit");
 	}
 	StrAllocCopy(jtp->file, JThead->file);
     }
@@ -156,7 +157,7 @@ PUBLIC BOOL LYJumpInit ARGS1 (char *, config)
      * Complete the initialization of config.
      */
     if (cp) {
-	jtp->key = remap(cp, "JUMP");	    /* key is present, (re)map it */
+	jtp->key = remap(cp, "JUMP", FALSE);  /* key is present, (re)map it */
 	cp = strtok(NULL, "\n");	    /* get prompt, if present */
 	if (cp && *cp)
 	    StrAllocCopy(jtp->msg, cp);     /* prompt is present, load it */
@@ -191,9 +192,10 @@ PUBLIC char *LYJump ARGS1(int, key)
     while (jtp && jtp->key && jtp->key != key)
 	jtp = jtp->next;
     if (!jtp) {
-	char msg[40];
-	sprintf(msg, KEY_NOT_MAPPED_TO_JUMP_FILE, key);
+	char *msg = 0;
+	HTSprintf0(&msg, KEY_NOT_MAPPED_TO_JUMP_FILE, key);
 	HTAlert(msg);
+	FREE(msg);
 	return NULL;
     }
     if (!jtp->table)
@@ -255,11 +257,8 @@ check_recall:
 	    HTUserMsg(RANDOM_URL_DISALLOWED);
 	    return NULL;
 	}
-	StrAllocCopy(temp, "Go ");
-	StrAllocCat(temp, bp);
-	strcpy(buf, temp);
-	FREE(temp);
-	return(bp=buf);
+	sprintf(buf, "Go %.*s", (int)sizeof(buf) - 4, bp);
+	return (bp = buf);
     }
 #endif /* PERMIT_GOTO_FROM_JUMP */
 
@@ -283,7 +282,7 @@ check_recall:
 	    ShortcutNum = 0;
 	if ((cp=(char *)HTList_objectAt(jtp->history,
 					ShortcutNum)) != NULL) {
-	    strcpy(buf, cp);
+	    LYstrncpy(buf, cp, sizeof(buf)-1);
 	    if (jump_buffer && jtp->shortcut &&
 		!strcmp(buf, jtp->shortcut)) {
 		_statusline(EDIT_CURRENT_SHORTCUT);
@@ -323,7 +322,7 @@ check_recall:
 	    ShortcutNum = ShortcutTotal - 1;
 	if ((cp=(char *)HTList_objectAt(jtp->history,
 					ShortcutNum)) != NULL) {
-	    strcpy(buf, cp);
+	    LYstrncpy(buf, cp, sizeof(buf)-1);
 	    if (jump_buffer && jtp->shortcut &&
 		!strcmp(buf, jtp->shortcut)) {
 		_statusline(EDIT_CURRENT_SHORTCUT);
@@ -350,7 +349,7 @@ check_recall:
 				 jtp->nel, sizeof(JumpDatum), LYCompare);
     if (!found) {
 	user_message("Unknown target '%s'", buf);
-	sleep(AlertSecs);
+	LYSleepAlert();
     }
 
     StrAllocCopy(jtp->shortcut, bp);
