@@ -10,9 +10,10 @@
 #include <LYUtils.h>
 #include <GridText.h>
 #include <LYList.h>
-#include <LYSignal.h>
+#include <LYClean.h>
 #include <LYGlobalDefs.h>
 #include <LYCharUtils.h>
+#include <LYCharSets.h>
 
 #ifdef DIRED_SUPPORT
 #include <LYUpload.h>
@@ -32,17 +33,6 @@
 **			Clear:	we only get addresses.
 */
 
-static char *list_filename = 0;
-
-/*
- *  Returns the name of the file used for the List Page, if one has
- *  been created, as a full URL; otherwise, returns an empty string.
- * - kw
- */
-PUBLIC char * LYlist_temp_url NOARGS
-{
-    return list_filename ? list_filename : "";
-}
 
 PUBLIC int showlist ARGS2(
 	document *,	newdoc,
@@ -50,9 +40,11 @@ PUBLIC int showlist ARGS2(
 {
     int cnt;
     int refs, hidden_links;
-    static char tempfile[256];
+    static char tempfile[LY_MAXPATH];
+    static BOOLEAN last_titles = TRUE;
     FILE *fp0;
     char *Address = NULL, *Title = NULL, *cp = NULL;
+    char *LinkTitle = NULL;  /* Rel stored as property of link, not of dest */
     BOOLEAN intern_w_post = FALSE;
     char *desc = "unknown field or link";
 
@@ -68,30 +60,42 @@ PUBLIC int showlist ARGS2(
 	return(-1);
     }
 
-    LYRemoveTemp(tempfile);
-    if ((fp0 = LYOpenTemp(tempfile, HTML_SUFFIX, "w")) == NULL) {
+    if (LYReuseTempfiles && titles == last_titles) {
+	fp0 = LYOpenTempRewrite(tempfile, HTML_SUFFIX, "w");
+    } else {
+	LYRemoveTemp(tempfile);
+	fp0 = LYOpenTemp(tempfile, HTML_SUFFIX, "w");
+    }
+    if (fp0 == NULL) {
 	HTUserMsg(CANNOT_OPEN_TEMP);
 	return(-1);
     }
 
-    LYLocalFileToURL(&list_filename, tempfile);
+    LYLocalFileToURL(&(newdoc->address), tempfile);
 
-    StrAllocCopy(newdoc->address, list_filename);
+    LYRegisterUIPage(newdoc->address,
+		     titles ? UIP_LIST_PAGE : UIP_ADDRLIST_PAGE);
+    last_titles = titles;
     LYforce_HTML_mode = TRUE;	/* force this file to be HTML */
     LYforce_no_cache = TRUE;	/* force this file to be new */
 
+#ifdef EXP_ADDRLIST_PAGE
+    if (titles != TRUE)
+	BeginInternalPage(fp0, ADDRLIST_PAGE_TITLE, LIST_PAGE_HELP);
+    else
+#endif
     BeginInternalPage(fp0, LIST_PAGE_TITLE, LIST_PAGE_HELP);
 
     StrAllocCopy(Address, HTLoadedDocumentURL());
     LYEntify(&Address, FALSE);
-    fprintf(fp0, "References in %s<p>\n",
-	((Address != NULL && *Address != '\0') ? Address : "this document:"));
+    fprintf(fp0, "%s%s<p>\n", gettext("References in "),
+	((Address != NULL && *Address != '\0') ? Address : gettext("this document:")));
     FREE(Address);
     if (refs > 0) {
 	fprintf(fp0, "<%s compact>\n", ((keypad_mode == NUMBERS_AS_ARROWS) ?
 				       "ol" : "ul"));
 	if (hidden_links > 0)
-	    fprintf(fp0, "<lh><em>Visible links:</em>\n");
+	    fprintf(fp0, "<lh><em>%s</em>\n", gettext("Visible links:"));
     }
     if (hidden_links > 0) {
 	if (LYHiddenLinks == HIDDENLINKS_IGNORE)
@@ -119,7 +123,7 @@ PUBLIC int showlist ARGS2(
 	     *	right in connection with always treating this file as
 	     *	HIDDENLINKS_MERGE in GridText.c - kw
 	     */
-	    if (keypad_mode == LINKS_AND_FORM_FIELDS_ARE_NUMBERED) {
+	    if (keypad_mode == LINKS_AND_FIELDS_ARE_NUMBERED) {
 		HText_FormDescNumber(cnt, (char **)&desc);
 		fprintf(fp0,
 		"<li><a id=%d href=\"#%d\">form field</a> = <em>%s</em>\n",
@@ -149,11 +153,21 @@ PUBLIC int showlist ARGS2(
 	}
 	address =  HTAnchor_address(dest);
 	title = titles ? HTAnchor_title(parent) : NULL;
+	if (dest_intl) {
+	    HTSprintf0(&LinkTitle, "(internal)");
+	} else if (titles && child->mainLink.type &&
+		   dest == child->mainLink.dest &&
+		   !strncmp(HTAtom_name(child->mainLink.type),
+			    "RelTitle: ", 10)) {
+	    HTSprintf0(&LinkTitle, "(%s)", HTAtom_name(child->mainLink.type)+10);
+	} else {
+	    FREE(LinkTitle);
+	}
 	StrAllocCopy(Address, address);
 	FREE(address);
 	LYEntify(&Address, TRUE);
 	if (title && *title) {
-	    StrAllocCopy(Title, title);
+	    LYformTitle(&Title, title);
 	    LYEntify(&Title, TRUE);
 	    if (*Title) {
 		cp = strchr(Address, '#');
@@ -164,7 +178,7 @@ PUBLIC int showlist ARGS2(
 
 	fprintf(fp0, "<li><a href=\"%s\"%s>%s%s%s%s%s</a>\n", Address,
 			dest_intl ? " TYPE=\"internal link\"" : "",
-			dest_intl ? "(internal) " : "",
+			LinkTitle ? LinkTitle : "",
 			((HTAnchor*)parent != dest) && Title ? "in " : "",
 			(char *)(Title ? Title : Address),
 			(Title && cp) ? " - " : "",
@@ -173,6 +187,7 @@ PUBLIC int showlist ARGS2(
 	FREE(Address);
 	FREE(Title);
     }
+    FREE(LinkTitle);
 
     if (hidden_links > 0) {
 	if (refs > 0)
@@ -181,7 +196,7 @@ PUBLIC int showlist ARGS2(
 							"ol" : "ul"));
 	fprintf(fp0, "<%s compact>\n", ((keypad_mode == NUMBERS_AS_ARROWS) ?
 					"ol continue" : "ul"));
-	fprintf(fp0, "<lh><em>Hidden links:</em>\n");
+	fprintf(fp0, "<lh><em>%s</em>\n", gettext("Hidden links:"));
     }
 
     for (cnt = 0; cnt < hidden_links; cnt++) {
@@ -237,13 +252,10 @@ PUBLIC void printlist ARGS2(
 	FILE *, 	fp,
 	BOOLEAN,	titles)
 {
-#ifdef VMS
-    extern BOOLEAN HadVMSInterrupt;
-#endif /* VMS */
     int cnt;
     int refs, hidden_links;
     char *address = NULL;
-    char *desc = "unknown field or link";
+    char *desc = gettext("unknown field or link");
 
     refs = HText_sourceAnchors(HTMainText);
     if (refs <= 0 && LYHiddenLinks != HIDDENLINKS_SEPARATE)
@@ -252,9 +264,9 @@ PUBLIC void printlist ARGS2(
     if (refs <= 0 && hidden_links <= 0) {
 	return;
     } else {
-	fprintf(fp, "\n%s\n\n", "References");
+	fprintf(fp, "\n%s\n\n", gettext("References"));
 	if (hidden_links > 0) {
-	    fprintf(fp, "   Visible links:\n");
+	    fprintf(fp, "   %s\n", gettext("Visible links"));
 	    if (LYHiddenLinks == HIDDENLINKS_IGNORE)
 		hidden_links = 0;
 	}
@@ -274,7 +286,7 @@ PUBLIC void printlist ARGS2(
 		 *  the list page match the numbering in the original document,
 		 *  but won't create a forward link to the form. - FM && LE
 		 */
-		if (keypad_mode == LINKS_AND_FORM_FIELDS_ARE_NUMBERED) {
+		if (keypad_mode == LINKS_AND_FIELDS_ARE_NUMBERED) {
 		    HText_FormDescNumber(cnt, (char **)&desc);
 		    fprintf(fp, "%4d. form field = %s\n", cnt, desc);
 		}
@@ -282,7 +294,7 @@ PUBLIC void printlist ARGS2(
 	    }
 	    dest = HTAnchor_followMainLink((HTAnchor *)child);
 	    /*
-	     *	Ignore if child anchor points to itself, i.e. we had
+	     *	Ignore if child anchor points to itself, i.e., we had
 	     *	something like <A NAME=xyz HREF="#xyz"> and it is not
 	     *	treated as a hidden link.  Useful if someone 'P'rints
 	     *	the List Page (which isn't a very useful action to do,
@@ -304,7 +316,7 @@ PUBLIC void printlist ARGS2(
 	}
 
 	if (hidden_links > 0)
-	    fprintf(fp, "%s   Hidden links:\n", ((refs > 0) ? "\n" : ""));
+	    fprintf(fp, "%s   %s\n", ((refs > 0) ? "\n" : ""), gettext("Hidden links:"));
 	for (cnt = 0; cnt < hidden_links; cnt++) {
 	    StrAllocCopy(address, HText_HiddenLinkAt(HTMainText, cnt));
 	    if (!(address && *address)) {
