@@ -40,6 +40,10 @@
 extern int default_fg;
 extern int default_bg;
 
+#if HAVE_USE_DEFAULT_COLORS && USE_DEFAULT_COLORS
+extern int lynx_default_colors NOPARAMS;
+#endif
+
 extern int check_color PARAMS((char * color, int the_default));
 #endif
 
@@ -47,6 +51,8 @@ extern void read_cfg PARAMS((char *cfg_filename, char *parent_filename, int nest
 extern void free_lynx_cfg NOPARAMS;
 extern BOOLEAN have_read_cfg;
 
-extern char *lynx_cfg_infopage NOPARAMS;
+extern int lynx_cfg_infopage PARAMS((document *newdoc));
+extern int lynx_compile_opts PARAMS((document *newdoc));
+extern void reload_read_cfg NOPARAMS; /* implemented in LYMain.c */
 
 #endif /* LYREADCFG_H */
