@@ -3,11 +3,11 @@
 #define LYSEARCH_H
 
 #ifndef LYSTRUCTS_H
-#include "LYStructs.h"
+#include <LYStructs.h>
 #endif /* LYSTRUCT_H */
 
 extern BOOL textsearch PARAMS((document *cur_doc,
-			       char *prev_target, BOOL next));
+			       char *prev_target, int target_size, BOOL next));
 
 #define IN_FILE 1
 #define IN_LINKS 2
