@@ -6,7 +6,7 @@
 **	Reread the upload menu page every time, in case the "upload" directory
 **	  has changed (make the current directory that for the upload process).
 **	Prompt for the upload file name if there is no "%s" in the command
-**	  string. Most protocols allow the user to specify the file name
+**	  string.  Most protocols allow the user to specify the file name
 **	  from the client side.  Xmodem appears to be the only that can't
 **	  figure out the filename from the transfer data so it needs the
 **	  information from lynx (or an upload script which prompts for it).
@@ -21,7 +21,6 @@
 #include <LYCurses.h>
 #include <LYUtils.h>
 #include <LYGlobalDefs.h>
-#include <LYSignal.h>
 #include <LYStrings.h>
 #include <LYClean.h>
 #include <LYGetFile.h>
@@ -31,7 +30,7 @@
 #include <LYexit.h>
 #include <LYLeaks.h>
 
-PUBLIC char LYUploadFileURL[LY_MAXPATH] = "\0";
+#define SUBDIR_COMMAND "cd %s ; "
 
 /*
  *  LYUpload uploads a file to a given location using a
@@ -42,19 +41,15 @@ PUBLIC char LYUploadFileURL[LY_MAXPATH] = "\0";
 PUBLIC int LYUpload ARGS1(
 	char *, 	line)
 {
-    char *method, *directory, *dir;
+    char *method, *directory;
     int method_number;
     int count;
+    char *the_upload = 0;
     char tmpbuf[LY_MAXPATH];
-    char buffer[LY_MAXPATH];
+    char *filename = NULL;
     lynx_html_item_type *upload_command = 0;
-    int c;
-    char *cp;
     FILE *fp;
-    char cmd[20 + (LY_MAXPATH*2)];
-#ifdef VMS
-    extern BOOLEAN HadVMSInterrupt;
-#endif /* VMS */
+    char *the_command = 0;
 
     /*
      *	Use configured upload commands.
@@ -81,14 +76,14 @@ PUBLIC int LYUpload ARGS1(
      *	Parsed out the Method and the Location?
      */
     if (upload_command->command == NULL) {
-	HTAlert("ERROR! - upload command is misconfigured");
+	HTAlert(gettext("ERROR! - upload command is misconfigured"));
 	goto failed;
     }
 
     /*
      *	Care about the local name?
      */
-    if (strstr(upload_command->command, "%s")) {
+    if (HTCountCommandArgs (upload_command->command)) {
 	/*
 	 *  Commands have the form "command %s [etc]"
 	 *  where %s is the filename.
@@ -103,101 +98,79 @@ retry:
 	    goto cancelled;
 
 	if (strstr(tmpbuf, "../") != NULL) {
-	    HTAlert("Illegal redirection \"../\" found! Request ignored.");
+	    HTAlert(gettext("Illegal redirection \"../\" found! Request ignored."));
 	    goto cancelled;
 	} else if (strchr(tmpbuf, '/') != NULL) {
-	    HTAlert("Illegal character \"/\" found! Request ignored.");
+	    HTAlert(gettext("Illegal character \"/\" found! Request ignored."));
 	    goto cancelled;
 	} else if (tmpbuf[0] == '~') {
-	    HTAlert("Illegal redirection using \"~\" found! Request ignored.");
+	    HTAlert(gettext("Illegal redirection using \"~\" found! Request ignored."));
 	    goto cancelled;
 	}
-	sprintf(buffer, "%s/%s", directory, tmpbuf);
+	HTSprintf0(&filename, "%s/%s", directory, tmpbuf);
 
-	if (no_dotfiles || !show_dotfiles) {
-	    if (*LYPathLeaf(buffer) == '.') {
-		HTAlert(FILENAME_CANNOT_BE_DOT);
-		_statusline(NEW_FILENAME_PROMPT);
-		goto retry;
-	    }
+#if HAVE_POPEN
+	if (LYIsPipeCommand(filename)) {
+	    HTAlert(CANNOT_WRITE_TO_FILE);
+	    _statusline(NEW_FILENAME_PROMPT);
+	    goto retry;
 	}
-
-	/*
-	 *  See if it already exists.
-	 */
-	if ((fp = fopen(buffer, "r")) != NULL) {
-	    fclose(fp);
-
-#ifdef VMS
-	    _statusline(FILE_EXISTS_HPROMPT);
-#else
-	    _statusline(FILE_EXISTS_OPROMPT);
-#endif /* VMS */
-	    c = 0;
-	    while (TOUPPER(c) != 'Y' && TOUPPER(c) != 'N' && c != 7 && c != 3)
-		c = LYgetch();
-#ifdef VMS
-	    if (HadVMSInterrupt) {
-		HadVMSInterrupt = FALSE;
-		goto cancelled;
-	    }
-#endif /* VMS */
-
-	    if (c == 7 || c == 3) { /* Control-G or Control-C */
-		goto cancelled;
-	    }
-
-	    if (TOUPPER(c) == 'N') {
-		_statusline(NEW_FILENAME_PROMPT);
-		goto retry;
-	    }
+#endif
+	switch (LYValidateOutput(filename)) {
+	case 'Y':
+	    break;
+	case 'N':
+	    goto retry;
+	default:
+	    goto cancelled;
 	}
 
 	/*
 	 *  See if we can write to it.
 	 */
-	CTRACE(tfp, "LYUpload: filename is %s", buffer);
+	CTRACE((tfp, "LYUpload: filename is %s", filename));
 
-	if ((fp = fopen(buffer, "w")) != NULL) {
+	if ((fp = fopen(filename, "w")) != NULL) {
 	    fclose(fp);
-	    remove(buffer);
+	    remove(filename);
 	} else {
 	    HTAlert(CANNOT_WRITE_TO_FILE);
 	    _statusline(NEW_FILENAME_PROMPT);
 	    goto retry;
 	}
 
-#if defined (VMS) || defined (__EMX__) || defined(__DJGPP__)
-	sprintf(tmpbuf, upload_command->command, buffer, "", "", "", "", "");
-#else
-	cp = quote_pathname(buffer); /* to prevent spoofing of the shell */
-	sprintf(tmpbuf, upload_command->command, cp, "", "", "", "", "");
-	FREE(cp);
-#endif /* VMS */
+	HTAddParam(&the_upload, upload_command->command, 1, filename);
+	HTEndParam(&the_upload, upload_command->command, 1);
     } else {			/* No substitution, no changes */
-	strcpy(tmpbuf, upload_command->command);
+	StrAllocCopy(the_upload, upload_command->command);
     }
 
-    dir = quote_pathname(directory);
-    sprintf(cmd, "cd %s ; %s", dir, tmpbuf);
-    FREE(dir);
+    HTAddParam(&the_command, SUBDIR_COMMAND, 1, directory);
+    HTEndParam(&the_command, SUBDIR_COMMAND, 1);
+    StrAllocCat(the_command, the_upload);
+
+    CTRACE((tfp, "command: %s\n", the_command));
+
     stop_curses();
-    CTRACE(tfp, "command: %s\n", cmd);
-    LYSystem(cmd);
+    LYSystem(the_command);
     start_curses();
+
+    FREE(the_command);
+    FREE(the_upload);
 #ifdef UNIX
-    chmod(buffer, HIDE_CHMOD);
+    if (filename != 0)
+	chmod(filename, HIDE_CHMOD);
 #endif /* UNIX */
-    /* don't remove(file); */
+    FREE(filename);
 
     return 1;
 
 failed:
-    HTAlert("Unable to upload file.");
+    HTAlert(gettext("Unable to upload file."));
     return 0;
 
 cancelled:
-    HTInfoMsg("Cancelling.");
+    HTInfoMsg(CANCELLING);
     return 0;
 }
 
@@ -218,8 +191,13 @@ PUBLIC int LYUpload_options ARGS2(
     static char curloc[LY_MAXPATH];
     char *cp;
 
-    LYRemoveTemp(tempfile);
-    if ((fp0 = LYOpenTemp(tempfile, HTML_SUFFIX, "w")) == NULL) {
+    if (LYReuseTempfiles) {
+	fp0 = LYOpenTempRewrite(tempfile, HTML_SUFFIX, "w");
+    } else {
+	LYRemoveTemp(tempfile);
+	fp0 = LYOpenTemp(tempfile, HTML_SUFFIX, "w");
+    }
+    if (fp0 == NULL) {
 	HTAlert(CANNOT_OPEN_TEMP);
 	return(-1);
     }
@@ -230,17 +208,17 @@ PUBLIC int LYUpload_options ARGS2(
     cp = HTfullURL_toFile(directory);
     strcpy(curloc,cp);
     LYTrimPathSep(curloc);
-    free(cp);
+    FREE(cp);
 #endif /* VMS */
 
     LYLocalFileToURL(newfile, tempfile);
-    strcpy(LYUploadFileURL, *newfile);
+    LYRegisterUIPage(*newfile, UIP_UPLOAD_OPTIONS);
 
     BeginInternalPage(fp0, UPLOAD_OPTIONS_TITLE, UPLOAD_OPTIONS_HELP);
 
     fprintf(fp0, "<pre>\n");
-    fprintf(fp0, "   <em>Upload To:</em> %s\n", curloc);
-    fputs("\nUpload options:\n", fp0);
+    fprintf(fp0, "   <em>%s</em> %s\n", gettext("Upload To:"), curloc);
+    fprintf(fp0, "\n%s\n", gettext("Upload options:"));
 
     if (uploaders != NULL) {
 	for (count = 0, cur_upload = uploaders;
@@ -249,7 +227,7 @@ PUBLIC int LYUpload_options ARGS2(
 	    fprintf(fp0, "   <a href=\"LYNXDIRED://UPLOAD=%d/TO=%s\">",
 			 count, curloc);
 	    fprintf(fp0, (cur_upload->name ?
-			  cur_upload->name : "No Name Given"));
+			  cur_upload->name : gettext("No Name Given")));
 	    fprintf(fp0, "</a>\n");
 	}
     } else {
