@@ -6,8 +6,8 @@
 #include <LYStructs.h>
 #endif /* LYSTRUCTS_H */
 
-extern void read_rc NOPARAMS;
-extern int save_rc NOPARAMS;
+extern void read_rc PARAMS((FILE *));
+extern int save_rc PARAMS((FILE *));
 
 #endif /* LYRCFILE_H */
 
