@@ -18,7 +18,6 @@
 
 #define DONT_USE_SOCKS5
 #include <HTUtils.h>
-#include <tcp.h>
 /*
  *  Don't try to use LYexit().
  */
@@ -26,6 +25,10 @@
 #undef exit
 #endif /* exit */
 
+#ifndef TOUPPER
+#define TOUPPER(c) (islower((unsigned char)c) ? toupper((unsigned char)c) : (c))
+#endif /* !TOLOWER */
+
 #ifndef TOLOWER
 #define TOLOWER(c) (isupper((unsigned char)c) ? tolower((unsigned char)c) : (c))
 #endif /* !TOLOWER */
@@ -40,25 +43,39 @@
  */
 typedef u16 unicode;
 
+static FILE *chdr = 0;
+
 /*
- * Since we're writing the formatted file to stdout, ensure that we flush
+ * Since we may be writing the formatted file to stdout, ensure that we flush
  * everything before leaving, since some old (and a few not-so-old) platforms
  * that do not implement POSIX 'exit()'.
  */
-#define done(code)  \
-    fflush(stdout); \
-    fflush(stderr); \
-    exit(code)
-
-PRIVATE void usage ARGS1(
-	char *,		argv0)
+PRIVATE void done PARAMS((int code)) GCC_NORETURN;
+PRIVATE void done ARGS1(int, code)
 {
-    fprintf(stderr, "Usage: \n");
-    fprintf(stderr,
-	    "        %s chartable [charsetmimename] [charsetdisplayname]\n",
-	    argv0);
-    fprintf(stderr,
-	    "Utility to convert .tbl into .h files for Lynx compilation.\n");
+    fflush(chdr);
+    fclose(chdr);
+    fflush(stderr);
+    exit(code);
+}
+
+PRIVATE void usage NOARGS
+{
+    static CONST char *tbl[] = {
+	"Usage: makeuctb [parameters]",
+	"",
+	"Utility to convert .tbl into .h files for Lynx compilation.",
+	"",
+	"Parameters (all are optional):",
+	"  1: the input file (normally {filename}.tbl, but \"-\" for stdin",
+	"  2: the output file (normally {filename}.tbl but \"-\" for stdout",
+	"  3: charset mime name",
+	"  4: charset display name"
+    };
+    unsigned n;
+    for (n = 0; n < TABLESIZE(tbl); n++) {
+	fprintf(stderr, "%s\n", tbl[n]);
+    };
     done(EX_USAGE);
 }
 
@@ -72,7 +89,7 @@ PUBLIC int strncasecomp ARGS3(
     CONST char *q = b;
 
     for (p = a, q = b; ; p++, q++) {
-        int diff;
+	int diff;
 	if (p == (a+n))
 	    return 0;	/*   Match up to n characters */
 	if (!(*p && *q))
@@ -91,7 +108,7 @@ PRIVATE int getunicode ARGS1(
 
     while (*p == ' ' || *p == '\t')
 	p++;
-	
+
     if (*p == '-') {
 	return -2;
     } else if (*p != 'U' || p[1] != '+' ||
@@ -109,26 +126,31 @@ PRIVATE int getunicode ARGS1(
 unicode unitable[MAX_FONTLEN][255];
 int unicount[MAX_FONTLEN];
 
-struct unimapdesc_str themap_str = {0, NULL};
+struct unimapdesc_str themap_str = {0, NULL, 0, 0};
 
-char *tblname;
+PRIVATE char *tblname;
+PRIVATE char *hdrname;
 
 PRIVATE int RawOrEnc = 0;
 PRIVATE int Raw_found = 0;		/* whether explicit R directive found */
+PRIVATE int CodePage = 0;
+PRIVATE int CodePage_found = 0;		/* whether explicit C directive found */
+
+#define MAX_UNIPAIRS 2500
 
 PRIVATE void addpair_str ARGS2(
 	char *,		str,
 	int,		un)
 {
-   int i;
+   int i = 0;
 
     if (un <= 0xfffe) {
 	if (!themap_str.entry_ct) {
 	    /*
 	     *  Initialize the map for replacement strings.
 	     */
-	    themap_str.entries =
-	  (struct unipair_str *) malloc (2000 * sizeof (struct unipair_str));
+	    themap_str.entries = (struct unipair_str *) malloc (MAX_UNIPAIRS
+				* sizeof (struct unipair_str));
 	    if (!themap_str.entries) {
 		fprintf(stderr,
 			"%s: Out of memory\n", tblname);
@@ -149,9 +171,10 @@ PRIVATE void addpair_str ARGS2(
 	/*
 	 *  Add to list.
 	 */
-	if (themap_str.entry_ct > 1999) {
+	if (themap_str.entry_ct > MAX_UNIPAIRS-1) {
 	    fprintf(stderr,
-		"ERROR: Only 2000 unicode replacement strings permitted!\n");
+		    "ERROR: Only %d unicode replacement strings permitted!\n",
+		    MAX_UNIPAIRS);
 	    done(EX_DATAERR);
 	}
 	themap_str.entries[themap_str.entry_ct].unicode = un;
@@ -167,7 +190,7 @@ PRIVATE void addpair ARGS2(
 {
     int i;
 
-    if (!Raw_found) {       /* enc not (yet) explicitly given with 'R' */
+    if (!Raw_found) {	/* enc not (yet) explicitly given with 'R' */
 	if (fp >= 128) {
 	    if (RawOrEnc != UCT_ENC_8BIT && RawOrEnc <= UCT_ENC_8859) {
 		if (fp < 160) {	/* cannot be 8859 */
@@ -217,16 +240,55 @@ PUBLIC int main ARGS2(
 	int,		argc,
 	char **,	argv)
 {
+    static char *first_ifdefs[] = {
+	"/*",
+	" * Compile-in this chunk of code unless we've turned it off specifically",
+	" * or in general (id=%s).",
+	" */",
+	"",
+	"#ifndef INCL_CHARSET%s",
+	"#define INCL_CHARSET%s 1",
+	"",
+	"/*ifdef NO_CHARSET*/",
+	"#ifdef  NO_CHARSET",
+	"#undef  NO_CHARSET",
+	"#endif",
+	"#define NO_CHARSET 0 /* force default to always be active */",
+	"",
+	"/*ifndef NO_CHARSET%s*/",
+	"#ifndef NO_CHARSET%s",
+	"",
+	"#if    ALL_CHARSETS",
+	"#define NO_CHARSET%s 0",
+	"#else",
+	"#define NO_CHARSET%s 1",
+	"#endif",
+	"",
+	"#endif /* ndef(NO_CHARSET%s) */",
+	"",
+	"#if NO_CHARSET%s",
+	"#define UC_CHARSET_SETUP%s /*nothing*/",
+	"#else"
+    };
+    static char *last_ifdefs[] = {
+	"",
+	"#endif /* NO_CHARSET%s */",
+	"",
+	"#endif /* INCL_CHARSET%s */"
+    };
+
     FILE *ctbl;
     char buffer[65536];
+    char outname[256];
+    unsigned n;
     int fontlen;
     int i, nuni, nent;
-    int fp0, fp1, un0, un1;
+    int fp0 = 0, fp1 = 0, un0, un1;
     char *p, *p1;
     char *tbuf, ch;
 
-    if (argc < 2 || argc > 4) {
-	usage(argv[0]);
+    if (argc < 2 || argc > 5) {
+	usage();
     }
 
     if (!strcmp(argv[1], "-")) {
@@ -240,6 +302,31 @@ PUBLIC int main ARGS2(
 	}
     }
 
+    if (argc > 2) {
+	if (!strcmp(argv[2], "-")) {
+	    chdr = stdout;
+	    hdrname = "stdout";
+	} else {
+	    hdrname = argv[2];
+	}
+    } else if (ctbl == stdin) {
+	chdr = stdout;
+	hdrname = "stdout";
+    } else {
+	strcpy(hdrname = outname, tblname);
+	if ((p = strrchr(outname, '.')) == 0)
+	    p = outname + strlen(outname);
+	strcpy(p, ".h");
+    }
+
+    if (chdr == 0) {
+	chdr = fopen(hdrname, "w");
+	if (!chdr) {
+	    perror(hdrname);
+	    done(EX_NOINPUT);
+	}
+    }
+
     /*
      *  For now we assume the default font is always 256 characters.
      */
@@ -307,7 +394,7 @@ PUBLIC int main ARGS2(
 		}
 		p++;
 		while (*p == ' ' || *p == '\t') {
-	  	    p++;
+		    p++;
 		}
 		RawOrEnc = strtol(p,0,10);
 		Raw_found = 1;
@@ -316,7 +403,7 @@ PUBLIC int main ARGS2(
 	    /*
 	     *  Is this the default table?
 	     */
- 	    case 'D':
+	    case 'D':
 		if (p[1] == 'e' || p[1] == 'E') {
 		    buffer[sizeof(buffer) - 1] = '\0';
 		    if (!strncasecomp(p, "Default", 7)) {
@@ -333,7 +420,7 @@ PUBLIC int main ARGS2(
 	    /*
 	     *  Is this the default table?
 	     */
- 	    case 'F':
+	    case 'F':
 		if (p[1] == 'a' || p[1] == 'A') {
 		    buffer[sizeof(buffer) - 1] = '\0';
 		    if (!strncasecomp(p, "FallBack", 8)) {
@@ -380,6 +467,24 @@ PUBLIC int main ARGS2(
 		}
 		this_LYNXcharset[i] = '\0';
 		continue;
+
+	    /*
+	     *  Codepage number.  Three or four digit code.
+	     */
+	    case 'C':
+		if (p[1] == 'o' || p[1] == 'O') {
+		    buffer[sizeof(buffer) - 1] = '\0';
+		    if (!strncasecomp(p, "CodePage", 8)) {
+			p += 8;
+		    }
+		}
+		p++;
+		while (*p == ' ' || *p == '\t') {
+		    p++;
+		}
+		CodePage = strtol(p,0,10);
+		CodePage_found = 1;
+		continue;
 	}
 
 	if (*p == 'U') {
@@ -456,7 +561,7 @@ PUBLIC int main ARGS2(
 		for (ch = *(++p); (ch = *p) != '\0'; p++, p1++) {
 		    if ((unsigned char)ch < 32 || ch == '\\' || ch == '\"' ||
 			(unsigned char)ch >= 127) {
-			sprintf(p1, "\\%.3o", (unsigned char)ch); 
+			sprintf(p1, "\\%.3o", (unsigned char)ch);
 #ifdef NOTDEFINED
 			fprintf(stderr, "%s\n", tbuf);
 #endif /* NOTDEFINED */
@@ -469,7 +574,7 @@ PUBLIC int main ARGS2(
 	    *p1 = '\0';
 	    for (i = un0; i <= un1; i++) {
 #ifdef NOTDEFINED
-		printf("U+0x%x:%s\n", i, tbuf); */
+		fprintf(chdr, "U+0x%x:%s\n", i, tbuf); */
 #endif /* NOTDEFINED */
 		addpair_str(tbuf,i);
 	    }
@@ -485,7 +590,7 @@ PUBLIC int main ARGS2(
 	if (p1 == p) {
 	    fprintf(stderr, "Bad input line: %s\n", buffer);
 	    done(EX_DATAERR);
-        }
+	}
 	p = p1;
 
 	while (*p == ' ' || *p == '\t') {
@@ -499,7 +604,7 @@ PUBLIC int main ARGS2(
 		done(EX_DATAERR);
 	    }
 	    p = p1;
-        } else {
+	} else {
 	    fp1 = 0;
 	}
 
@@ -541,7 +646,7 @@ PUBLIC int main ARGS2(
 		    fprintf(stderr,
 			    " there should be a Unicode range.\n");
 		    done(EX_DATAERR);
-	        }
+		}
 		p++;
 		un1 = getunicode(&p);
 		if (un0 < 0 || un1 < 0) {
@@ -549,7 +654,7 @@ PUBLIC int main ARGS2(
      "%s: Bad Unicode range corresponding to font position range 0x%x-0x%x\n",
 			    tblname, fp0, fp1);
 		    done(EX_DATAERR);
-	        }
+		}
 		if (un1 - un0 != fp1 - fp0) {
 		    fprintf(stderr,
 			"%s: Unicode range U+%x-U+%x not of the same length",
@@ -558,7 +663,7 @@ PUBLIC int main ARGS2(
 			    " as font position range 0x%x-0x%x\n",
 			    fp0, fp1);
 		    done(EX_DATAERR);
-	        }
+		}
 		for (i = fp0; i <= fp1; i++) {
 		    addpair(i,un0-fp0+i);
 		}
@@ -608,7 +713,7 @@ PUBLIC int main ARGS2(
      *  Okay, we hit EOF, now output tables.
      */
     fclose(ctbl);
-  
+
 
     /*
      *  Compute total size of Unicode list.
@@ -618,8 +723,8 @@ PUBLIC int main ARGS2(
 	nuni += unicount[i];
     }
 
-    if (argc >= 3) {
-	strncpy(this_MIMEcharset,argv[2],UC_MAXLEN_MIMECSNAME);
+    if (argc > 3) {
+	strncpy(this_MIMEcharset,argv[3],UC_MAXLEN_MIMECSNAME);
     } else if (this_MIMEcharset[0] == '\0') {
 	strncpy(this_MIMEcharset,tblname,UC_MAXLEN_MIMECSNAME);
 	if ((p = strchr(this_MIMEcharset,'.')) != 0) {
@@ -629,8 +734,8 @@ PUBLIC int main ARGS2(
     for (p = this_MIMEcharset; *p; p++) {
 	*p = TOLOWER(*p);
     }
-    if (argc >= 4) {
-	strncpy(this_LYNXcharset,argv[3],UC_MAXLEN_LYNXCSNAME);
+    if (argc > 4) {
+	strncpy(this_LYNXcharset,argv[4],UC_MAXLEN_LYNXCSNAME);
     } else if (this_LYNXcharset[0] == '\0') {
 	strncpy(this_LYNXcharset,this_MIMEcharset,UC_MAXLEN_LYNXCSNAME);
     }
@@ -650,8 +755,8 @@ PUBLIC int main ARGS2(
 	this_isDefaultMap = !strncmp(this_MIMEcharset,"iso-8859-1", 10);
     }
     fprintf(stderr,
-    	    "makeuctb: %s: %stranslation map",
- 	    this_MIMEcharset, (this_isDefaultMap ? "default " : ""));
+	    "makeuctb: %s: %stranslation map",
+	    this_MIMEcharset, (this_isDefaultMap ? "default " : ""));
     if (this_isDefaultMap == 1) {
 	*id_append = '\0';
     } else {
@@ -664,7 +769,12 @@ PUBLIC int main ARGS2(
     }
     fprintf(stderr, " (%s).\n", id_append);
 
-    printf("\
+    for (n = 0; n < TABLESIZE(first_ifdefs); n++) {
+	fprintf(chdr, first_ifdefs[n], id_append);
+	fprintf(chdr, "\n");
+    }
+
+    fprintf(chdr, "\n\
 /*\n\
  *  uni_hash.tbl\n\
  *\n\
@@ -674,20 +784,20 @@ PUBLIC int main ARGS2(
  *\n\
  */\n\
 \n\
-static u8 dfont_unicount%s[%d] = \n\
+static CONST u8 dfont_unicount%s[%d] = \n\
 {\n\t", argv[0], argv[1], id_append, fontlen);
 
     for (i = 0; i < fontlen; i++) {
 	if (i >= 128 && unicount[i] > 0 && i < lowest_eight) {
 	    lowest_eight = i;
 	}
-	printf("%3d", unicount[i]);
+	fprintf(chdr, "%3d", unicount[i]);
 	if (i == (fontlen - 1)) {
-	    printf("\n};\n");
+	    fprintf(chdr, "\n};\n");
 	} else if ((i % 8) == 7) {
-	    printf(",\n\t");
+	    fprintf(chdr, ",\n\t");
 	} else {
-	    printf(", ");
+	    fprintf(chdr, ", ");
 	}
     }
 
@@ -700,10 +810,10 @@ static u8 dfont_unicount%s[%d] = \n\
     }
 
     if (nuni) {
-	printf("\nstatic u16 dfont_unitable%s[%d] = \n{\n\t",
-	       id_append, nuni);
+	fprintf(chdr, "\nstatic CONST u16 dfont_unitable%s[%d] = \n{\n\t",
+		id_append, nuni);
     } else {
-	printf("\nstatic u16 dfont_unitable%s[1]; /* dummy */\n", id_append);
+	fprintf(chdr, "\nstatic CONST u16 dfont_unitable%s[1]; /* dummy */\n", id_append);
     }
 
     fp0 = 0;
@@ -713,57 +823,62 @@ static u8 dfont_unicount%s[%d] = \n\
 	    fp0++;
 	    nent = 0;
 	}
-	printf("0x%04x", unitable[fp0][nent++]);
+	fprintf(chdr, "0x%04x", unitable[fp0][nent++]);
 	if (i == (nuni - 1)) {
-	    printf("\n};\n");
+	    fprintf(chdr, "\n};\n");
 	} else if ((i % 8) == 7) {
-	    printf(",\n\t");
+	    fprintf(chdr, ",\n\t");
 	} else {
-	    printf(", ");
+	    fprintf(chdr, ", ");
 	}
     }
 
     if (themap_str.entry_ct) {
-	printf("\n\
+	fprintf(chdr, "\n\
 static struct unipair_str repl_map%s[%d] = \n\
 {\n\t", id_append, themap_str.entry_ct);
     } else {
-	printf("\n\
+	fprintf(chdr, "\n\
 /* static struct unipair_str repl_map%s[]; */\n", id_append);
     }
-    
+
     for (i = 0; i < themap_str.entry_ct; i++) {
-	printf("{0x%x,\"%s\"}",
-	       themap_str.entries[i].unicode,
-	       themap_str.entries[i].replace_str);
+	fprintf(chdr, "{0x%x,\"%s\"}",
+		themap_str.entries[i].unicode,
+		themap_str.entries[i].replace_str);
 	if (i == (themap_str.entry_ct - 1)) {
-	    printf("\n};\n");
+	    fprintf(chdr, "\n};\n");
 	} else if ((i % 4) == 3) {
-	    printf(",\n\t");
+	    fprintf(chdr, ",\n\t");
 	} else {
-	    printf(", ");
+	    fprintf(chdr, ", ");
 	}
     }
     if (themap_str.entry_ct) {
-	printf("\n\
-static struct unimapdesc_str dfont_replacedesc%s = {%d,repl_map%s,",
+	fprintf(chdr, "\n\
+static CONST struct unimapdesc_str dfont_replacedesc%s = {%d,repl_map%s,",
 id_append, themap_str.entry_ct, id_append);
     } else {
-	printf("\n\
-static struct unimapdesc_str dfont_replacedesc%s = {0,NULL,",id_append);
+	fprintf(chdr, "\n\
+static CONST struct unimapdesc_str dfont_replacedesc%s = {0,NULL,",id_append);
     }
-    printf("%d,%d};\n",
-    this_isDefaultMap ? 1 : 0,
-    (useDefaultMap && !this_isDefaultMap) ? 1 : 0
+    fprintf(chdr, "%d,%d};\n",
+	    this_isDefaultMap ? 1 : 0,
+	    (useDefaultMap && !this_isDefaultMap) ? 1 : 0
     );
 
 
-    printf("#define UC_CHARSET_SETUP%s UC_Charset_Setup(\
+    fprintf(chdr, "#define UC_CHARSET_SETUP%s UC_Charset_Setup(\
 \"%s\",\\\n\"%s\",\\\n\
 dfont_unicount%s,dfont_unitable%s,%d,\\\n\
-dfont_replacedesc%s,%d,%d)\n",
+dfont_replacedesc%s,%d,%d,%d)\n",
 id_append, this_MIMEcharset, this_LYNXcharset,
-id_append, id_append, nuni, id_append, lowest_eight, RawOrEnc);
+id_append, id_append, nuni, id_append, lowest_eight, RawOrEnc, CodePage);
+
+    for (n = 0; n < TABLESIZE(last_ifdefs); n++) {
+	fprintf(chdr, last_ifdefs[n], id_append);
+	fprintf(chdr, "\n");
+    }
 
     done(EX_OK);
 }
